module mgdem
use mkinds
use mncio
use mdates
use mutil
!use mpobs 

integer(i4),parameter         :: nclon=1440,nclat=689,ncz=78,nct=14
real(r4),save                 :: clon(nclon),clat(nclat),clim_depths(ncz),ctime(0:nct-1)
real(r4)                      :: cv1(1,1,ncz,2),cv2(2,2,ncz,2),cv(2,2,ncz),util(ncz),cvar(ncz)
character(:),allocatable      :: gdem_data_location
real(r4),allocatable          :: ctemp(:,:,:,:),csaln(:,:,:,:),ctstd(:,:,:,:),csstd(:,:,:,:)

contains


subroutine initialize_gdem()        
implicit none

integer(i4)               :: i,j,k
real(r4)                  :: lnmn,lnmx,ltmn,ltmx,grd_size

clim_depths=[0, 2, 4, 6, 8, 10, 15, 20, 25, 30,35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 100, 110, 120, 130, 140, 150, 160, 170, 180, 190,&
       200, 220, 240, 260, 280, 300, 350, 400, 500, 600, 700, 800, 900, 1000, 1100, 1200, 1300, 1400, 1500, 1600, 1800, 2000, 2200, 2400, 2600, 2800, 3000, 3200, 3400, 3600, 3800,&
       4000, 4200, 4400, 4600, 4800,5000, 5200,5400, 5600, 5800, 6000, 6200, 6400, 6600]

lnmn=-180.0
lnmx=179.75
ltmn=-82.0
ltmx=89.0
grd_size=0.25

do i=1,nclon
  clon(i)=lnmn+(i-1)*grd_size
enddo

do j=1,nclat
  clat(j)=ltmn+(j-1)*grd_size
enddo

do i=0,nct-1
ctime(i)=i
enddo

allocate(ctemp(nclon,nclat,ncz,2))
allocate(csaln(nclon,nclat,ncz,2))
allocate(ctstd(nclon,nclat,ncz,2))
allocate(csstd(nclon,nclat,ncz,2))

end subroutine initialize_gdem


subroutine get_clim_fld(gdem_data_location,var,cdtg,dst_lon,dst_lat,nx,ny,cfld)
implicit none
character(*)              :: gdem_data_location,var,cdtg
integer(i4)               :: nx,ny
real(r4)                  :: cfld(nx,ny),dst_lon(nx,ny),dst_lat(nx,ny)
integer(i4)               :: iyyyy,imm,idd,ihh
real(r4)                  :: t1,t2,grdi,grdj,w1,w2,r,s
integer(i4)               :: ix,iy,iz,it
integer(i4)               :: fid,i,j,iunit,k
character(:), allocatable :: filename,fname
character(:), allocatable :: vname,cfile
integer(i4)               :: vtype
integer(i4)               :: vdims(2),vdim(1)
integer(i4),parameter     :: src_nx=1440,src_ny=689
real(r4), allocatable     :: util(:,:,:,:),src_fld(:,:)
logical(bl)               :: lmask(src_nx,src_ny),lperiodic
integer(i4)               :: iw(src_nx,src_ny)
character(cl)             :: buffer,depthFile

call c2idate(cdtg,iyyyy,imm,idd,ihh)

imm=imm+1

if(idd<=15) then
  t1=imm-1
  t2=imm

  w2=(15.5+idd)/31.0
  w1=(15.5-idd)/31.0

elseif(idd>15) then
  t1=imm
  t2=imm+1

  w1=(15.5+idd)/31.0
  w2=(15.5-idd)/31.0

endif

it=int(t1)

allocate(util(src_nx,src_ny,1,2),src_fld(src_nx,src_ny))


if (var=="ptem") cfile=trim(gdem_data_location)//"gdem_ptem.nc"
if (var=="saln") cfile=trim(gdem_data_location)//"gdem_saln.nc"
if (var=="tstd") cfile=trim(gdem_data_location)//"gdem_tstd.nc"
if (var=="sstd") cfile=trim(gdem_data_location)//"gdem_sstd.nc"

call nciopn(cfile,fid)
call nciorv(cfile,fid,var,util(:,:,:,1),[1,1,1,it],[src_nx,src_ny,1,2])
call nciorv(cfile,fid,var,util(:,:,:,2),[1,1,1,it],[src_nx,src_ny,1,2])
call nciocl(cfile,fid)

src_fld=w1*util(:,:,1,1)+w2*util(:,:,1,2)

call intp_fld(nclon,nclat,nx,ny,clon,clat,dst_lon,dst_lat,src_fld,cfld)



end subroutine

subroutine get_clim_profile(var,ptlon,ptlat,cdtg,npres,pres,ivar)

implicit none
character(*) :: var,cdtg
real(r4)     :: ptlon,ptlat


! internal variables

character(:),allocatable  :: cfile
integer(i4)               :: npres
integer(i4)   :: iyyyy,imm,idd,ihh,fid
real(r4)      :: ivar(npres),pres(npres)
real(r4)      :: t1,t2,grdi,grdj,w1,w2,r,s
integer(i4)   :: ix,iy,iz,it
integer(i4)   :: i,j,k

call c2idate(cdtg,iyyyy,imm,idd,ihh)
imm=imm+1
if(idd<=15) then 
  t1=imm-1
  t2=imm

  w2=(15.5+idd)/31.0
  w1=(15.5-idd)/31.0

elseif(idd>15) then 
  t1=imm
  t2=imm+1

    w1=(15.5+idd)/31.0
    w2=(15.5-idd)/31.0

endif  

!print *, idd,t1,t2,w1,w2,w1+w2
grdi=ridx(clon,ptlon)
grdj=ridx(clat,ptlat)
ix=int(grdi)
iy=int(grdj)
it=int(t1)
r=grdi-ix
s=grdj-iy


if (var=="ptem") cfile=trim(gdem_data_location)//"gdem_ptem.nc"
if (var=="saln") cfile=trim(gdem_data_location)//"gdem_saln.nc"
if (var=="tstd") cfile=trim(gdem_data_location)//"gdem_tstd.nc"
if (var=="sstd") cfile=trim(gdem_data_location)//"gdem_sstd.nc"

call nciopn(cfile,fid)
if(grdi>2. .and. grdi<nclon-2 .and. grdj>2. .and. grdj<nclat-2) then
    
        ! do a bilinear intp
        call nciorv(cfile,fid,var,cv2,[ix,iy,1,it],[2,2,ncz,2])
        cv(:,:,:)=w1*cv2(:,:,:,1)+w2*cv2(:,:,:,2)
        do k=1,ncz
          util(k) = (1. - s)*((1. - r)*cv(1,1,k) + r*cv(2,1,k)) + s*((1. -r)*cv(1,2,k) + r*cv(2,2,k))
        enddo
else
        call nciorv(cfile,fid,var,cv1,[ix,iy,1,it],[1,1,ncz,2])
        util=w1*cv1(1,1,:,1)+w2*cv1(1,1,:,2)
endif
call nciocl(cfile,fid)

! final interp to target vertical grid
call akspline(ncz,clim_depths,util,npres,pres,ivar)

end subroutine


subroutine get_clim_profiles(ncount,nplon,nplat,cdtg,npres,pres,atemp,asaln,atstd,asstd)

implicit none
character(*)              :: cdtg
character(:),allocatable  :: cfile
integer(i4)               :: ncount,npres
integer(i4)               :: iyyyy,imm,idd,ihh,fid
real(r4)                  :: nplon(ncount),nplat(ncount),pres(npres)
real(r4)                  :: atemp(ncount,npres),asaln(ncount,npres),atstd(ncount,npres),asstd(ncount,npres)
real(r4)                  :: t1,t2,grdi,grdj,w1,w2,r,s
integer(i4)               :: ix,iy,iz,it
integer(i4)               :: i,j,k

call c2idate(cdtg,iyyyy,imm,idd,ihh)
imm=imm+1
if(idd<=15) then
  t1=imm-1
  t2=imm

  w2=(15.5+idd)/31.0
  w1=(15.5-idd)/31.0

elseif(idd>15) then
  t1=imm
  t2=imm+1

    w1=(15.5+idd)/31.0
    w2=(15.5-idd)/31.0

endif

it=int(t1)

! load ctemp
cfile=trim(gdem_data_location)//"gdem_ptem.nc"
call nciopn(cfile,fid)
call nciorv(cfile,fid,"ptem",ctemp,[1,1,1,it],[nclon,nclat,ncz,2])
call nciocl(cfile,fid)

!load csaln
cfile=trim(gdem_data_location)//"gdem_saln.nc"
call nciopn(cfile,fid)
call nciorv(cfile,fid,"saln",csaln,[1,1,1,it],[nclon,nclat,ncz,2])
call nciocl(cfile,fid)

!load tstd
cfile=trim(gdem_data_location)//"gdem_tstd.nc"
call nciopn(cfile,fid)
call nciorv(cfile,fid,"tstd",ctstd,[1,1,1,it],[nclon,nclat,ncz,2])
call nciocl(cfile,fid)
!load sstd
cfile=trim(gdem_data_location)//"gdem_sstd.nc"
call nciopn(cfile,fid)
call nciorv(cfile,fid,"sstd",csstd,[1,1,1,it],[nclon,nclat,ncz,2])
call nciocl(cfile,fid)


do i=1,ncount

      grdi=ridx(clon,nplon(i))
      grdj=ridx(clat,nplat(i))
      ix=int(grdi)
      iy=int(grdj)
      it=int(t1)
      r=grdi-ix
      s=grdj-iy

        if(grdi>2. .and. grdi<nclon-2 .and. grdj>2. .and. grdj<nclat-2) then
              ! do a bilinear intp
                 ! temp
                 cv(:,:,:)=w1*ctemp(ix:ix+1,iy:iy+1,:,1)+w2*ctemp(ix:ix+1,iy:iy+1,:,2)
                 do k=1,ncz
                   util(k) = (1. - s)*((1. - r)*cv(1,1,k) + r*cv(2,1,k)) + s*((1. -r)*cv(1,2,k) + r*cv(2,2,k))
                 enddo
                 call akspline(ncz,clim_depths,util,npres,pres,atemp(i,:))

                 ! saln
                 cv(:,:,:)=w1*csaln(ix:ix+1,iy:iy+1,:,1)+w2*csaln(ix:ix+1,iy:iy+1,:,2)
                 do k=1,ncz
                   util(k) = (1. - s)*((1. - r)*cv(1,1,k) + r*cv(2,1,k)) + s*((1. -r)*cv(1,2,k) + r*cv(2,2,k))
                 enddo
                 call akspline(ncz,clim_depths,util,npres,pres,asaln(i,:)) 

                 ! tstd
                 cv(:,:,:)=w1*ctstd(ix:ix+1,iy:iy+1,:,1)+w2*ctstd(ix:ix+1,iy:iy+1,:,2)
                 do k=1,ncz
                   util(k) = (1. - s)*((1. - r)*cv(1,1,k) + r*cv(2,1,k)) + s*((1. -r)*cv(1,2,k) + r*cv(2,2,k))
                 enddo
                 call akspline(ncz,clim_depths,util,npres,pres,atstd(i,:))
        
                 ! saln
                 cv(:,:,:)=w1*csstd(ix:ix+1,iy:iy+1,:,1)+w2*csstd(ix:ix+1,iy:iy+1,:,2)
                 do k=1,ncz
                   util(k) = (1. - s)*((1. - r)*cv(1,1,k) + r*cv(2,1,k)) + s*((1. -r)*cv(1,2,k) + r*cv(2,2,k))
                 enddo
                 call akspline(ncz,clim_depths,util,npres,pres,asstd(i,:))
               
          else
                 !temp  
                 util=w1*ctemp(ix,iy,:,1)+w2*ctemp(ix,iy,:,2)
                 call akspline(ncz,clim_depths,util,npres,pres,atemp(i,:))
                 
                 !saln  
                 util=w1*csaln(ix,iy,:,1)+w2*csaln(ix,iy,:,2)
                 call akspline(ncz,clim_depths,util,npres,pres,asaln(i,:))

                 !tstd  
                 util=w1*ctstd(ix,iy,:,1)+w2*ctstd(ix,iy,:,2)
                 call akspline(ncz,clim_depths,util,npres,pres,atstd(i,:))

                 !sstd
                 util=w1*csstd(ix,iy,:,1)+w2*csstd(ix,iy,:,2)
                 call akspline(ncz,clim_depths,util,npres,pres,asstd(i,:))

       endif
    enddo
end subroutine






end module      
