MODULE mhycom
USE mkinds
USE mseawater
USE mgrid, only : sigver,mdepth

INTERFACE readFromBlkdat
          MODULE PROCEDURE intFromBlkdat
          MODULE PROCEDURE realFromBlkdat
END INTERFACE readFromBlkdat

CONTAINS


SUBROUTINE readarchv(archvFile,nx,ny,nz,fssh,fmld,fuba,fvba,fuvl,fvvl,fthk,ftem,fsal,fden,lsteric)
USE mkinds
USE mabio

!--------------------------
! version 0.1 -- 2013/03/02
! ashwanth srinivasan
! Tendral LLC, Key Biscayne
! -------------------------
! This subroutine reads state variables from the hycom archv files in .ab format.
! The variables and the order are listed in the .b summary file. Here is a 
! sample from hycom version 2.2.35:
!
!montg1   =   11670048  40521.000  2 34.000  -7.3649502E+00   7.1277499E+00
!srfhgt   =   11670048  40521.000  0  0.000  -4.4968357E+00   6.7205215E+00
!surflx   =   11670048  40521.000  0  0.000  -3.2207242E+02   2.7023532E+02
!salflx   =   11670048  40521.000  0  0.000  -4.8381768E-02   4.9601831E-03
!bl_dpth  =   11670048  40521.000  0  0.000   2.6522498E+04   2.0802259E+06
!mix_dpth =   11670048  40521.000  0  0.000   1.5019707E+04   2.0520280E+06
!u_btrop  =   11670048  40521.000  0  0.000  -8.7710834E-01   1.0526524E+00
!v_btrop  =   11670048  40521.000  0  0.000  -6.8449908E-01   1.2685642E+00
!u-vel.   =   11670048  40521.000  1 27.100  -1.6910708E+00   1.7091720E+00
!v-vel.   =   11670048  40521.000  1 27.100  -2.3331254E+00   1.8231095E+00
!thknss   =   11670048  40521.000  1 27.100   2.9418000E+04   2.9418000E+04
!temp     =   11670048  40521.000  1 27.100   1.5441038E+01   2.9824617E+01
!salin    =   11670048  40521.000  1 27.100   1.3702556E+01   3.7636162E+01
!u-vel.   =   11670048  40521.000  2 28.100  -1.6836554E+00   1.6924214E+00
!v-vel.   =   11670048  40521.000  2 28.100  -2.2934167E+00   1.8132000E+00
!thknss   =   11670048  40521.000  2 28.100   3.4713238E+04   3.4713238E+04
!temp     =   11670048  40521.000  2 28.100   1.5408322E+01   2.9761263E+01
!salin    =   11670048  40521.000  2 28.100   1.3694813E+01   3.7636604E+01
!........
!........
!Check to make sure that the fields and the order of reading in this subroutine
!are are consistent with the appropriate .b file

IMPLICIT NONE

! Argument Declarations

CHARACTER(len=*), INTENT(IN)           :: archvFile
INTEGER(i4),      INTENT(IN)           :: nx
INTEGER(i4),      INTENT(IN)           :: ny
INTEGER(i4),      INTENT(IN)           :: nz
REAL(r4),         INTENT(INOUT)        :: fssh(:,:)
REAL(r4),         INTENT(INOUT)        :: fmld(:,:)
REAL(r4),         INTENT(INOUT)        :: fuba(:,:)
REAL(r4),         INTENT(INOUT)        :: fvba(:,:)
REAL(r4),         INTENT(INOUT)        :: fuvl(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fvvl(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fthk(:,:,:)
REAL(r4),         INTENT(INOUT)        :: ftem(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fsal(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fden(:,:,:)
Logical(bl)                            :: lsteric
! Local Variables

INTEGER(i4)                            :: iunit
INTEGER(i4)                            :: k,i,j
CHARACTER(:),ALLOCATABLE               :: FileName
REAL(r4)                               :: hmin
REAL(r4)                               :: hmax


! Executable Section

FileName=archvFile(1:LEN_TRIM(archvFile))
iunit=11

! Initialize hycom style IO

CALL zaiost(nx,ny)
CALL zaiopn(FileName,iunit)

! First read 2d fields skipping over some fields 

CALL zaiosk(iunit)                      ! skip montg
CALL zaiord(iunit,fssh(:,:),hmin,hmax)  ! read ssh
if(lsteric) then
CALL zaiosk(iunit)                      ! skip steric 
endif
CALL zaiosk(iunit)                      ! skip surflx  
CALL zaiosk(iunit)                      ! skip salflx
CALL zaiosk(iunit)                      ! skip bl_dpth
CALL zaiord(iunit,fmld(:,:),hmin,hmax)  ! read mx_dpth
CALL zaiord(iunit,fuba(:,:),hmin,hmax)  ! read ubaro
CALL zaiord(iunit,fvba(:,:),hmin,hmax)  ! read vbaro

! Next read layered fields

do k=1,nz

CALL zaiord(iunit,fuvl(:,:,k),hmin,hmax) ! read xward velocities
CALL zaiord(iunit,fvvl(:,:,k),hmin,hmax) ! read yward velocities
CALL zaiord(iunit,fthk(:,:,k),hmin,hmax) ! read thickness
CALL zaiord(iunit,ftem(:,:,k),hmin,hmax) ! read pot. temperature
CALL zaiord(iunit,fsal(:,:,k),hmin,hmax) ! read salinity

enddo

! Close .ab files
call zaiocl(iunit)
fden=0.0
END SUBROUTINE


SUBROUTINE readarchm(archvFile,nx,ny,nz,fssh,fmld,fuba,fvba,fuvl,fvvl,fthk,ftem,fsal,fden)
USE mkinds
USE mabio

!--------------------------
! version 0.1 -- 2013/03/02
! ashwanth srinivasan
! Tendral LLC, Key Biscayne
! -------------------------
! This subroutine reads state variables from the hycom archm files in .ab format.
! The variables and the order are listed in the .b summary file. Here is a 
! sample from hycom version 2.2.35:
!
!montg1   =        360  40371.001  0  0.000  -1.3490833E+00   3.0635638E+00
!srfhgt   =        360  40371.999  0  0.000  -9.5571166E-01   4.0864615E+00
!surflx   =        360  40371.500  0  0.000  -1.2940276E+02   1.9590085E+02
!salflx   =        360  40371.500  0  0.000  -1.2873916E-01   6.7397580E-02
!bl_dpth  =        360  40371.500  0  0.000   2.7769725E+04   8.1792475E+05
!mix_dpth =        360  40371.500  0  0.000   1.5703066E+04   7.9142381E+05
!tmix     =        360  40371.500  0  0.000   2.5674660E+01   3.0157867E+01
!smix     =        360  40371.500  0  0.000   2.6779840E+00   3.6732365E+01
!thmix    =        360  40371.500  0 34.000  -2.7410286E+01  -1.3730425E+00
!umix     =        360  40371.500  0  0.000  -1.4978151E+00   9.4613039E-01
!vmix     =        360  40371.500  0  0.000  -1.1516929E+00   1.2597382E+00
!kemix    =        360  40371.500  0 34.000   8.1536682E-06   1.2460749E+00
!u_btrop  =        360  40371.500  0  0.000  -8.8034350E-01   2.9426852E-01
!v_btrop  =        360  40371.500  0  0.000  -4.4977045E-01   7.4802178E-01
!kebtrop  =        360  40371.500  0  0.000   5.3182987E-08   4.5221969E-01
!u-vel.   =        360  40371.500  1 27.100  -1.4978151E+00   9.4613039E-01
!v-vel.   =        360  40371.500  1 27.100  -1.1516929E+00   1.2597382E+00
!k.e.     =        360  40371.500  1 27.100   8.1536682E-06   1.2460749E+00
!thknss   =        360  40371.500  1 27.100   2.9417807E+04   2.9418262E+04
!temp     =        360  40371.500  1 27.100   2.5674660E+01   3.0157867E+01
!salin    =        360  40371.500  1 27.100   2.6779840E+00   3.6732365E+01
!density  =        360  40371.500  1 27.100  -2.7410286E+01  -1.3730425E+00
!........
!........
!Check to make sure that the fields and the order of reading in this subroutine
!are are consistent with the appropriate .b file. Note this is different than
!the one for archv files -- contains additional mixed layer fields, mean KE and
!density fields. 

IMPLICIT NONE

! Argument Declarations

CHARACTER(len=*), INTENT(IN)           :: archvFile
INTEGER(i4),      INTENT(IN)           :: nx
INTEGER(i4),      INTENT(IN)           :: ny
INTEGER(i4),      INTENT(IN)           :: nz
REAL(r4),         INTENT(INOUT)        :: fssh(:,:)
REAL(r4),         INTENT(INOUT)        :: fuba(:,:)
REAL(r4),         INTENT(INOUT)        :: fvba(:,:)
REAL(r4),         INTENT(INOUT)        :: fmld(:,:)
REAL(r4),         INTENT(INOUT)        :: fuvl(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fvvl(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fthk(:,:,:)
REAL(r4),         INTENT(INOUT)        :: ftem(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fsal(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fden(:,:,:)

! Local Variables

INTEGER(i4)                            :: iunit
INTEGER(i4)                            :: k
CHARACTER(:),ALLOCATABLE               :: FileName
REAL(r4)                               :: hmin
REAL(r4)                               :: hmax


! Executable Section

FileName=archvFile(1:LEN_TRIM(archvFile))
iunit=11

! Initialize hycom style IO

CALL zaiost(nx,ny)
CALL zaiopn(FileName,iunit)

! First read 2d fields skipping over some fields 

CALL zaiosk(iunit)                      ! skip montg
CALL zaiord(iunit,fssh(:,:),hmin,hmax)  ! read ssh
CALL zaiosk(iunit)                      ! skip surflx  
CALL zaiosk(iunit)                      ! skip salflx
CALL zaiosk(iunit)                      ! skip bl_dpth
!CALL zaiosk(iunit)                      ! skip mx_dpth
CALL zaiord(iunit,fmld(:,:),hmin,hmax)  ! read mx_dpth
CALL zaiosk(iunit)                      ! skip tmix
CALL zaiosk(iunit)                      ! skip smix
CALL zaiosk(iunit)                      ! skip thmix
CALL zaiosk(iunit)                      ! skip umix
CALL zaiosk(iunit)                      ! skip vmix
CALL zaiosk(iunit)                      ! skip kemix
CALL zaiord(iunit,fuba(:,:),hmin,hmax)  ! read ubaro
CALL zaiord(iunit,fvba(:,:),hmin,hmax)  ! read vbaro
CALL zaiosk(iunit)                      ! skip kebaro

! Next read layered fields

do k=1,nz

CALL zaiord(iunit,fuvl(:,:,k),hmin,hmax) ! read xward velocities
CALL zaiord(iunit,fvvl(:,:,k),hmin,hmax) ! read yward velocities
CALL zaiosk(iunit)                       ! skip ke
CALL zaiord(iunit,fthk(:,:,k),hmin,hmax) ! read thickness
CALL zaiord(iunit,ftem(:,:,k),hmin,hmax) ! read pot. temperature
CALL zaiord(iunit,fsal(:,:,k),hmin,hmax) ! read salinity
CALL zaiord(iunit,fden(:,:,k),hmin,hmax) ! read pot. density
!CALL zaiosk(iunit)                       ! skip ke

enddo

! Close .ab files
call zaiocl(iunit)

END SUBROUTINE

SUBROUTINE readrst(rstFile,nx,ny,nz,kapref,pref,thbase,thref,thflag,imask,fssh,fsst,fsss,&
                fmld,fuba,fvba,fpba,fuvl,fvvl,fthk,ftem,fsal,fden,fthkk,fpsikk,fmgp,fpin,ltides)
USE mkinds
USE mabio
USE mutil
USE mseawater

!--------------------------
! version 0.1 -- 2013/03/08
! ashwanth srinivasan
! Tendral LLC, Key Biscayne
! -------------------------
! This subroutine reads state variables from the hycom restart files in .ab format.
! The variables and the order are listed in the .b summary file. Here is a 
! sample from hycom version 2.2.35:
! RESTART2: iexpt,iversn,yrflag,sigver =    100    22     3     2
! RESTART2: nstep,dtime,thbase =     14745240   40959.0000000000 34.0000000000000
! ...... followed by.......
! 2 timelevels for uvl 2*nz records
! 2 timelevels for vvv 2*nz records
! 2 timelevels for thk 2*nz records
! 2 timelevels for tem 2*nz records
! 2 timelevels for sal 2*nz records
! ..... folowed by 2d variables .....
! ubavg   : layer,tlevel,range =   0  1    -8.3196521E-01   2.8939241E-01
! ubavg   : layer,tlevel,range =   0  2    -8.3003163E-01   2.8942749E-01
! ubavg   : layer,tlevel,range =   0  3    -8.3184856E-01   2.8939536E-01
! vbavg   : layer,tlevel,range =   0  1    -2.7728680E-01   9.4853359E-01
! vbavg   : layer,tlevel,range =   0  2    -2.7770573E-01   9.5031315E-01
! vbavg   : layer,tlevel,range =   0  3    -2.7730376E-01   9.4855434E-01
! pbavg   : layer,tlevel,range =   0  1    -3.3379419E+03   5.6284883E+03
! pbavg   : layer,tlevel,range =   0  2    -3.3375940E+03   5.6294312E+03
! pbavg   : layer,tlevel,range =   0  3    -3.3379399E+03   5.6285479E+03
! pbot    : layer,tlevel,range =   0  1     0.0000000E+00   5.5346384E+07
! psikk   : layer,tlevel,range =   0  1    -1.9182270E+01   3.5327896E-02
! thkk    : layer,tlevel,range =   0  1    -1.5279036E+01   3.1667187E+00
! dpmixl  : layer,tlevel,range =   0  1     1.5899357E+04   1.1576816E+06
! dpmixl  : layer,tlevel,range =   0  2     1.5899357E+04   1.1576816E+06
! Note: Depending on the thermobaric reference state psikk and thkk may have 2
! states. Check to make sure that the fields and the order of reading in this 
! subroutine are are consistent with the appropriate .b file. We always read
! the second time level, change this below if necessary. Also note that when
! reading restart files we need to calculate the surface elevation and can 
! store forecast Montgomery potential for calculating geostrophic adjustment
! after analysis. 

IMPLICIT NONE

! Argument Declarations

CHARACTER(len=*), INTENT(IN)           :: rstFile
INTEGER(i4),      INTENT(IN)           :: nx
INTEGER(i4),      INTENT(IN)           :: ny
INTEGER(i4),      INTENT(IN)           :: nz
INTEGER(i4),      INTENT(IN)           :: kapref
REAL(r4),         INTENT(IN)           :: pref
REAL(r4),         INTENT(IN)           :: thbase
REAL(r4),         INTENT(IN)           :: thref
INTEGER(i4),      INTENT(IN)           :: thflag
integer(i4),      INTENT(IN)           :: imask(:,:)
REAL(r4),         INTENT(INOUT)        :: fssh(:,:)
REAL(r4),         INTENT(INOUT)        :: fsst(:,:)
REAL(r4),         INTENT(INOUT)        :: fsss(:,:)
REAL(r4),         INTENT(INOUT)        :: fmld(:,:)
REAL(r4),         INTENT(INOUT)        :: fuba(:,:)
REAL(r4),         INTENT(INOUT)        :: fvba(:,:)
REAL(r4),         INTENT(INOUT)        :: fuvl(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fvvl(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fthk(:,:,:)
REAL(r4),         INTENT(INOUT)        :: ftem(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fsal(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fden(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fpin(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fmgp(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fpsikk(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fthkk(:,:,:)
REAL(r4),         INTENT(INOUT)        :: fpba(:,:)
logical(bl),       INTENT(IN)           :: ltides
! Local Variables
INTEGER(i4)                            :: iunit
INTEGER(i4)                            :: k,i,j
CHARACTER(:),ALLOCATABLE               :: FileName
REAL(r4)                               :: hmin
REAL(r4)                               :: hmax
REAL(r4),  ALLOCATABLE                 :: ftmp(:,:,:)
! Executable Section



FileName=rstFile(1:LEN_TRIM(rstFile))
iunit=11

! Initialize hycom style IO

CALL zaiost(nx,ny)
CALL zaiopn(FileName,iunit)

! skip nz fields and read nz uvl(xward) fields at the second time level 

DO k=1,nz
  CALL zaiosk(iunit)                    
ENDDO

DO k=1,nz
  CALL zaiord(iunit,fuvl(:,:,k),hmin,hmax)
ENDDO

! skip nz fields and read nz vvl(yward) fields at the second time level 

DO k=1,nz
  CALL zaiosk(iunit)                      
ENDDO

DO k=1,nz
   CALL zaiord(iunit,fvvl(:,:,k),hmin,hmax) 
ENDDO

! skip nz fields and read nz thickness fields at the second time level 

DO k=1,nz
  CALL zaiosk(iunit)
ENDDO

DO k=1,nz
   CALL zaiord(iunit,fthk(:,:,k),hmin,hmax) 
ENDDO

! skip nz fields and read nz tem fields at the second time level 

DO k=1,nz
  CALL zaiosk(iunit)
ENDDO

DO k=1,nz
  CALL zaiord(iunit,ftem(:,:,k),hmin,hmax)   
ENDDO

! skip nz fields and read nz sal fields at the second time level 

DO k=1,nz
  CALL zaiosk(iunit)
ENDDO

DO k=1,nz
  CALL zaiord(iunit,fsal(:,:,k),hmin,hmax)  
ENDDO

! skip uhrly and vhrly
if(ltides) then
!!Alex DO k=1,4*nz
DO k=1,49*2
  CALL zaiosk(iunit)
ENDDO
endif

! now read 2d fields; get the mid-time level
! ubaro 

CALL zaiosk(iunit) 
CALL zaiord(iunit,fuba(:,:),hmin,hmax) ! mid-time ubaro
CALL zaiosk(iunit)

! vbaro 

CALL zaiosk(iunit)
CALL zaiord(iunit,fvba(:,:),hmin,hmax) ! mid-time vbaro
CALL zaiosk(iunit)

! pbaro

CALL zaiosk(iunit)
CALL zaiord(iunit,fpba(:,:),hmin,hmax) ! mid-time pbaro
CALL zaiosk(iunit)

! skip pbot

CALL zaiosk(iunit)

! read psikk and thkk depending on the thermobaric ref state

IF (kapref >= 0) THEN
    CALL zaiord(iunit,fpsikk(:,:,1),hmin,hmax) ! psikk
    CALL zaiord(iunit,fthkk(:,:,1),hmin,hmax) ! thkk
ELSE

    CALL zaiord(iunit,fpsikk(:,:,1),hmin,hmax) ! psikk - ref state 1    
    CALL zaiord(iunit,fpsikk(:,:,2),hmin,hmax) ! psikk - ref state 2
    CALL zaiord(iunit,fthkk(:,:,1),hmin,hmax)  ! thkk  - ref state 1 
    CALL zaiord(iunit,fthkk(:,:,2),hmin,hmax)  ! thkk  - ref state 2  

ENDIF

! skip first dpbl and read second

CALL zaiosk(iunit)
CALL zaiord(iunit,fmld(:,:),hmin,hmax) ! dpbl 

! Close .ab files
call zaiocl(iunit)

! calculate density

do  k=1,nz
 do j=1,ny
  do i=1,nx
    if(mdepth(i,j)<2.00**99) then
    fden(i,j,k)=sig(ftem(i,j,k),fsal(i,j,k),sigver)
    else
    fden(i,j,k)=0.0
    fsal(i,j,k)=0.0
    ftem(i,j,k)=0.0
    fthk(i,j,k)=0.0    
    endif
  enddo
 enddo
enddo

do j=1,ny
  do i=1,nx
    if(mdepth(i,j)>2.00**99) then
      fpba(i,j)=0.0
      fpsikk(i,j,:)=0.0
      fthkk(i,j,:)=0.0
    endif
  enddo
enddo



! calculate ssh
CALL getssh(nx,ny,nz,kapref,pref,thbase,thref,imask,fsal,ftem,fthk,fden-thbase,fpba,fpsikk,fthkk,fmgp,fssh)
where(mdepth>2.00**99) fssh=0.0
!print *, minval(fssh),maxval(fssh)
! calculate sst and sss and fpin

fsst=ftem(:,:,1)
fsss=fsal(:,:,1)
!print *, minval(fsst),maxval(fsst)
!print *, minval(fsss),maxval(fsss)
!fpin=2.00**100
fpin(:,:,1)=0.0
do k=1,nz
   fpin(:,:,k+1)= fpin(:,:,k)+fthk(:,:,k)/9806.
!   print *, minval(fpin(:,:,k)),maxval(fpin(:,:,k))
enddo


END SUBROUTINE

subroutine getssh(nx,ny,nz,kapref,pref,thbase,thref,imask,s,t,dp,r,pb,psikk,thkk,fmontg,fssh)
use mkinds
use mncio
use mseawater

implicit none

integer(i4), intent(in)         :: nx
integer(i4), intent(in)         :: ny
integer(i4), intent(in)         :: nz
integer(i4), intent(in)         :: kapref
real(r4),    intent(in)         :: pref
real(r4),    intent(in)         :: thbase
real(r4),    intent(in)         :: thref
integer(i4),  intent(in)        :: imask(:,:)
real(r4),    intent(in)         :: pb(:,:)
real(r4),    intent(in)         :: psikk(:,:,:)
real(r4),    intent(in)         :: thkk(:,:,:)
real(r4),    intent(in)         :: s(:,:,:)
real(r4),    intent(in)         :: t(:,:,:)
real(r4),    intent(in)         :: dp(:,:,:)
real(r4),    intent(in)         :: r(:,:,:)
real(r4),    intent(out)        :: fssh(:,:)
real(r4),    intent(out)        :: fmontg(:,:,:)
real(r4),allocatable            :: montgl(:,:,:,:)


real(r4),    allocatable        :: thstar(:,:)
real(r4),    allocatable        :: montg(:,:,:)
real(r4),    allocatable        :: montg1(:,:)
real(r4),    allocatable        :: p(:,:,:)
integer(i4), allocatable        :: kapi(:,:)
real(r4),    allocatable        :: skap(:,:)
real(r4),    allocatable        :: tbaric(:,:)
real(r4)                        :: oneta
real(r4)                        :: hmin
real(r4)                        :: hmax
integer(i4)                     :: i
integer(i4)                     :: j
integer(i4)                     :: k
integer(i4)                     :: kk
integer(i4)                     :: ifunit
character(:),allocatable        :: FileName
integer(i4)                     :: fid



allocate(thstar(nz,2))
allocate(montg(nx,ny,2))
allocate(montg1(nx,ny))
allocate(p(nx,ny,nz+1))
allocate(montgl(nx,ny,nz,2))

kk=nz

! set intf pressure.

p(:,:,1)=0.0
do k=1,nz
 do j=1,ny
  do i=1,nx
   p(i,j,k+1)=p(i,j,k)+dp(i,j,k) 
  enddo
 enddo 
enddo

!read tbaric if needed depending on kapref
if(kapref == -1) then
      allocate(skap(nx,ny))
      allocate(kapi(nx,ny))
      allocate(tbaric(0:nx+1,0:ny+1))
      filename="gridinfo.nc"
      CALL nciopn(filename,fid)
      CALL nciorv(filename,fid,"tbaric",tbaric(1:nx,1:ny))
      CALL nciocl(filename,fid)
      
      tbaric(0,:)=tbaric(1,:)
      tbaric(:,0)=tbaric(:,1)
      tbaric(nx+1,:)=tbaric(nx,:)
      tbaric(:,ny+1)=tbaric(:,ny)

      where(tbaric>3.0) tbaric=1.0 
      do j= 1,ny
        do i= 1,nx
          if (imask(i,j) .eq. 1) then
             if (max(tbaric(i,j),tbaric(i,j),tbaric(i+1,j),tbaric(i,j-1),tbaric(i,j+1) ).gt.2.0) then
                  kapi(i,j) = 3              !kapi
                  skap(i,j) = 3.0 - tbaric(i,j)
             else
                  kapi(i,j) = 1              !kapi
                  skap(i,j) = tbaric(i,j) - 1.0
            endif
          endif 
         enddo
      enddo
endif

do j = 1,ny
 do i =1,nx
   if (imask(i,j) .eq. 1) then
    do k=1,nz
      if(kapref.ne.0) then !thermobaric case
   
        if(kapref.gt.0) then
             thstar(k,1) = r(i,j,k)+kappaf(t(i,j,k),s(i,j,k),r(i,j,k)+thbase,p(i,j,k),pref,kapref)
        else
             thstar(k,1) = r(i,j,k)+kappaf(t(i,j,k),s(i,j,k),r(i,j,k)+thbase,p(i,j,k),pref,2)
             thstar(k,2) = r(i,j,k)+kappaf(t(i,j,k),s(i,j,k),r(i,j,k)+thbase,p(i,j,k),pref,kapi(i,j))
        endif

    else ! non thermobaric case

        thstar(k,1) =   r(i,j,k)

    endif
      !print *, thstar(k,1),thstar(k,2) 
     enddo


oneta = 1.0 + pb(i,j)/p(i,j,kk+1)
!    print *, oneta,p(i,j,kk+1)
if (kapref.eq.-1) then
    montg(i,j,1) =  psikk(i,j,1) + (p(i,j,kk+1)*(thkk(i,j,1)-thstar(kk,1))-pb(i,j)*(thstar(kk,1)+thbase))*thref**2
    montg(i,j,2) =  psikk(i,j,2) + (p(i,j,kk+1)*(thkk(i,j,2)-thstar(kk,2))-pb(i,j)*(thstar(kk,2)+thbase))*thref**2
    montgl(i,j,kk,1)=montg(i,j,1)
    montgl(i,j,kk,2)=montg(i,j,2)
else
    montg(i,j,1) =  psikk(i,j,1) +  (p(i,j,kk+1)*(thkk(i,j,1)-thstar(kk,1))-pb(i,j)*(thstar(kk,1)+thbase))*thref**2
    montgl(i,j,kk,1)=montg(i,j,1)
endif !kapref.eq.-1
!    print *, montgl(i,j,kk,1),psikk(i,j,1),thkk(i,j,1)
do k=kk-1,1,-1
    montg(i,j,1) = montg(i,j,1)+p(i,j,k+1)*oneta*(thstar(k+1,1)-thstar(k,1))*thref**2
    montgl(i,j,k,1)=montg(i,j,1)
!    print *, montgl(i,j,k,1)
    if(kapref.eq.-1) then
       montg(i,j,2) = montg(i,j,2)+p(i,j,k+1)*oneta*(thstar(k+1,2)-thstar(k,2))*thref**2
       montgl(i,j,k,2)=montg(i,j,2)
    endif 
!print *,k, montgl(i,j,k,1),montgl(i,j,k,2)
enddo


if(kapref .ge. 0) then
   montg1(i,j) = montg(i,j,1)
   fmontg(i,j,:)=montgl(i,j,:,1)
else
   montg1(i,j) = skap(i,j)*montg(i,j,1) + (1.0-skap(i,j))*montg(i,j,2)
   fmontg(i,j,:)=skap(i,j)*montgl(i,j,:,1)+(1.0-skap(i,j))*montgl(i,j,:,2)
   
endif

   fssh(i,j) = (montg1(i,j) + thref*pb(i,j))/9.8
 !  print *, fssh(i,j),thref*pb(i,j),montg1(i,j),skap(i,j)
 else
   fssh(i,j)=2.00**100

  endif
 enddo
enddo
end subroutine getssh
!deep    z-level spacing minimum thickness (m)

SUBROUTINE read_hybrid_min_thickness(nz,ns,dp0k,ds0k)
IMPLICIT NONE

! Argument Declarations

INTEGER(i4),          INTENT(IN)      :: nz
INTEGER(i4),          INTENT(IN)      :: ns
REAL(r4)           ,  INTENT(OUT)     :: dp0k(nz)
REAL(r4)           ,  INTENT(OUT)     :: ds0k(ns)

! Local Variables

CHARACTER(LEN=cl)                     :: buffer
INTEGER(i4), PARAMETER                :: IACHAR_SPACE = 32
INTEGER(i4)                           :: i,j
INTEGER(i4)                           :: indx,mxindx
INTEGER(I4)                           :: IACHAR_Character
j=0
!try dp0k
OPEN(UNIT=20,FILE="blkdat.input",FORM="formatted",STATUS="old",ACCESS="stream")
DO WHILE (.true.)
   READ (20, '(A)', end=99) buffer
   buffer=ADJUSTL(buffer)
   indx = index (buffer,"dp0k")
   IF(indx .ne. 0) THEN
      mxindx=max(0,indx)
      j=j+1
      DO i = 1, LEN(buffer)

      ! Convert the current character to its position in the ASCII collating sequence

      IACHAR_Character = IACHAR(buffer(i:i))

      ! Find the index of first whitespace ' ' 
      IF ( IACHAR_Character == IACHAR_SPACE) THEN
               READ(buffer(1:i),*) dp0k(j)
      END IF

      END DO
    END IF
ENDDO
99 continue
CLOSE(20)

!try ds0k
j=0
OPEN(UNIT=20,FILE="blkdat.input",FORM="formatted",STATUS="old",ACCESS="stream")
DO WHILE (.true.)
   READ (20, '(A)', end=100) buffer
   buffer=ADJUSTL(buffer)
   indx = index (buffer,"ds0k")
   IF(indx .ne. 0) THEN
      mxindx=max(0,indx)
      j=j+1
      DO i = 1, LEN(buffer)

      ! Convert the current character to its position in the ASCII collating sequence

      IACHAR_Character = IACHAR(buffer(i:i))

      ! Find the index of first whitespace ' ' 
      IF ( IACHAR_Character == IACHAR_SPACE) THEN
               READ(buffer(1:i),*) ds0k(j)
      END IF

      END DO
    END IF
ENDDO
100 continue
CLOSE(20)

IF (mxindx == 0) then

dp0k=0.0
ds0k=0.0
ENDIF

END SUBROUTINE

SUBROUTINE readTargetDensities(nz,tsigma)
IMPLICIT NONE

! Argument Declarations

INTEGER(i4),          INTENT(IN)      :: nz
REAL(r4)           ,  INTENT(OUT)     :: tsigma(nz)

! Local Variables

CHARACTER(LEN=cl)                     :: buffer
INTEGER(i4), PARAMETER                :: IACHAR_SPACE = 32
INTEGER(i4)                           :: i,j
INTEGER(i4)                           :: indx
INTEGER(I4)                           :: IACHAR_Character

j=0
OPEN(UNIT=20,FILE="blkdat.input",FORM="formatted",STATUS="old",ACCESS="stream")
DO WHILE (.true.)
   READ (20, '(A)', end=99) buffer
   buffer=ADJUSTL(buffer)
   indx = index (buffer,"isopycnal target density")
   IF(indx .ne. 0) THEN
      j=j+1
      DO i = 1, LEN(buffer)
      
      ! Convert the current character to its position in the ASCII collating sequence

      IACHAR_Character = IACHAR(buffer(i:i))

      ! Find the index of first whitespace ' ' 
      IF ( IACHAR_Character == IACHAR_SPACE) THEN
               READ(buffer(1:i),*) tsigma(j)
      END IF

      END DO
    END IF
ENDDO
99 continue
close(20)
END SUBROUTINE

SUBROUTINE intFromBlkdat(vname,val)

IMPLICIT NONE

! Argument Declarations

CHARACTER(LEN=*), INTENT(IN)      :: vname
INTEGER           ,  INTENT(OUT)   :: val

! Local Variables

CHARACTER(LEN=80)                 :: buffer
INTEGER, PARAMETER                :: IACHAR_SPACE = 32
INTEGER                           :: i
INTEGER                           :: indx
INTEGER                           :: IACHAR_Character

OPEN(UNIT=20,FILE="blkdat.input",FORM="formatted",STATUS="old",ACCESS="stream")
rewind(20)
Outer :DO WHILE (.true.)
   READ (20, '(A)', end=99) buffer
   buffer=ADJUSTL(buffer)
   indx = index (buffer,vname)
   IF(indx .ne. 0) THEN
      DO i = 1, LEN(buffer)

      ! Convert the current character to its position in the ASCII collating sequence

      IACHAR_Character = IACHAR(buffer(i:i))

      ! Find the index of first whitespace ' ' 
      IF ( IACHAR_Character == IACHAR_SPACE) THEN
               READ(buffer(1:i),*) val
               EXIT outer
      END IF

      END DO
    END IF
ENDDO outer
99 continue
CLOSE(20)
END SUBROUTINE


SUBROUTINE realFromBlkdat(vname,val)

IMPLICIT NONE

! Argument Declarations

CHARACTER(LEN=*), INTENT(IN)      :: vname
REAL           ,  INTENT(OUT)     :: val

! Local Variables

CHARACTER(LEN=80)                 :: buffer
INTEGER, PARAMETER                :: IACHAR_SPACE = 32
INTEGER                           :: i
INTEGER                           :: indx
INTEGER                           :: IACHAR_Character

OPEN(UNIT=20,FILE="blkdat.input",FORM="formatted",STATUS="old",ACCESS="stream")
Outer :DO WHILE (.true.)
   READ (20, '(A)', end=99) buffer
   buffer=ADJUSTL(buffer)
   indx = index (buffer,vname)
   IF(indx .ne. 0) THEN
      DO i = 1, LEN(buffer)
      
      ! Convert the current character to its position in the ASCII collating sequence

      IACHAR_Character = IACHAR(buffer(i:i))

      ! Find the index of first whitespace ' ' 
      IF ( IACHAR_Character == IACHAR_SPACE) THEN
               READ(buffer(1:i),*) val
               EXIT outer
      END IF
              
      END DO
    END IF 
ENDDO outer
99 continue
CLOSE(20)
END SUBROUTINE


SUBROUTINE lilo(start,count,mask2d,ists,mdepth,cdepth,thref,thbase,dp0k,dpin,th3d,dpmixl,deltassh,dpout)

implicit none

INTEGER(i4)         :: start(3)
INTEGER(i4)         :: count(3)

integer(i4)         :: mask2d(count(1),count(2))
REAL(r4)            :: mdepth(count(1),count(2))
REAL(r4)            :: ists(count(1),count(2))
real(r4)            :: dpin(count(1),count(2),count(3))
real(r4)            :: th3d(count(1),count(2),count(3))
real(r4)            :: deltassh(count(1),count(2))
real(r4)            :: dpmixl(count(1),count(2))
real(r4)            :: dpout(count(1),count(2),count(3))
real(r4)            :: dp0k(count(3))
real(r4)            :: cdepth
real(r4)            :: thbase
real(r4)            :: thref

! Local variables

integer(i4)         :: i,j,k,k0,kk,kmn,kmx,err,fid
integer(i4)         :: k2,kk2,kml,kmn2,kmx2

real(r4)            :: ddp(count(1),count(2),count(3))
real(r4)            :: pin(count(1),count(2),count(3)+1)
!real(r4)            :: dp0k(count(3))
integer(i4)         :: called(count(1),count(2))
integer(i4)         :: error(count(1),count(2))

real(r4)            :: dz1d,dzout
real(r4)            :: den1d(count(3))
real(r4)            :: dpmin1d(count(3))
real(r4)            :: dp1d(count(3))
real(r4)            :: ddp1d(count(3))

real(r4)            :: den1d2(count(3))
real(r4)            :: dpmin1d2(count(3))
real(r4)            :: dp1d2(count(3))
real(r4)            :: ddp1d2(count(3))


real(r4),parameter  :: flag=2.00**100,onem=9806.0,qonem = 1.0/9806.,g =9.806
real(r4)            :: smin,smax,sumdp,fract

kk=count(3)
!calculate interface pressures and density

pin(:,:,1)=0.0
do k=1,kk
   pin(:,:,k+1)= pin(:,:,k)+dpin(:,:,k)
enddo

dpout=dpin


!convert to pressure units

dp0k=dp0k*onem

!loop for CH

DO j= 1,count(2)
 DO i= 1,count(1)
     called(i,j)=0.
     error(i,j)=1
! Adjust only if depth is greater than critical depth
     if(abs(deltassh(i,j)) > 0.02 .and. ists(i,j)>0.0 ) then
!     if (mask2d(i,j) .eq. 1 .and.mdepth(i,j).gt. cdepth)then

! all layers beneath and including the layer containing the mixed layer base
! will be adjusted if necessary. To extend the adjustment into all fixed thickness layers at and beneath the
! mixed layer base, all layers between the layer containing the base (kml) and
! the first layer significantly thicker than the specified minimum thickness
! (kmn) are consolidated into a single layer before the call to ch. After the
! ch call, all layers within this consolidated layer are thinned or thickened
! in the same proportion as the consolidated layer. k0 is set to 4 to exclude layers 1-3 from adjustment because they are
! assumed to always be in the mixed layer.

    k0=4
    kml=k0
    do k=k0,kk
        if(pin(i,j,k).gt.dpmixl(i,j))then
           kml=k
           exit
        endif
     enddo

    kmn=k0
    do k=k0,kk
       if(dpin(i,j,k).gt.dp0k(k)+onem)then
          kmn=k
          exit
       endif
    enddo

    kml=min(kml,kmn)

    kmx=kmn
    
    do k=kk,kmn,-1
       if(dpin(i,j,k).ge.onem.and.dpin(i,j,k).gt.dp0k(k) .and. pin(i,j,k).lt.mdepth(i,j)*onem) then
         kmx=k
         exit
       endif
    enddo


    do k=1,kk
        dz1d=deltassh(i,j)
        den1d(k)=(1/thref)/(1.0-thref*(th3d(i,j,k)+thbase))
        dpmin1d(k)=min(dpin(i,j,k),dp0k(k)) !in pressure unit
        dp1d(k)=dpin(i,j,k)  !in pressure units
    enddo



! consolidate layers from kml to kmn so that minimum thickness layers are also
! adjusted

    kk2=kk-kmn+kml
    kmn2=kml
    kmx2=kmx+kk2-kk

    den1d2=0.0
    dpmin1d2=0.0
    dp1d2=0.0
    ddp1d2=0.0

    do k=1,kk2
       if (k<kml) then
             den1d2(k)=den1d(k)
             dpmin1d2(k)=dpmin1d(k)
             dp1d2(k)=dp1d(k)
       elseif (k==kml) then
             dpmin1d2(k)=dpmin1d(k+kk-kk2)
           do k2=k,k+kmn-kml
             den1d2(k)=den1d2(k)+den1d(k2)*dp1d(k2)
             dp1d2(k)=dp1d2(k)+dp1d(k2)
           enddo
             den1d2(k)=den1d2(k)/dp1d2(k)
       elseif (k>kml) then
             den1d2(k)=den1d(k+kk-kk2)
             dpmin1d2(k)=dpmin1d(k+kk-kk2)
             dp1d2(k)=dp1d(k+kk-kk2)
       endif
     enddo

!sanity check before entering the adjustment process

    if(kmn2.ge.kmx2.or.kmn2.ge.kk2.or.kmx2.eq.1) then
       error(i,j)=1
    else
       called(i,j)=1
       CALL ch(kk2,kmn2,kmx2,dz1d,dzout,den1d2,g,dpmin1d2,dp1d2,ddp1d2,err)
     
       
! reset vertical arrays including those in the consolidated layer

       fract=(dp1d2(kml)+ddp1d2(kml))/dp1d2(kml)
       do k=1,kk2
          if (k<kml) then
             ddp1d(k)=ddp1d2(k)
          elseif (k==kml) then
             do k2=k,k+kmn-kml
                ddp1d(k2)=(fract-1.0)*dp1d(k2)
             enddo
          elseif (k>kml) then
             ddp1d(k+kk-kk2)=ddp1d2(k)
          endif
       enddo

!apply correction to pressure field
       if(err.eq.0)then
          error(i,j)=0
          do k=1,kk
             !ddp(i,j,k)=ddp1d(k)
             if(dpin(i,j,k)+ddp1d(k).gt.0.0 .and. ddp1d(k).ne.0.0) then
                 dpout(i,j,k)=dpin(i,j,k)+ddp1d(k)
                 !print *, k,ddp1d(k)/9806.,dpin(i,j,k)/9806.,dpout(i,j,k)/9806.
             endif    
          enddo
       endif ! err
       endif !called
       endif !ip=1

    enddo ! i loop
  enddo   ! j loop

END SUBROUTINE

subroutine CH(n_lyr, kmn, kmx, dzin, dzout, den, g, dpmin, dp, ddp,err)
use mkinds

! DESCRIPTION:  Computes innovations for layer (baroclinic pressure)
!               thickness from innovations for surface height using
!               Cooper-Haines method for HYCOM layers.
!
!               Whenever possible, only two layers will have
!               non-zero innovations, one above the main thermocline
!               and one below.
!               Their innovations are equal in magnitude and opposite
!               in sign, so that the resulting pressure thickness for
!               the water column will not change the bottom pressure.
!               The magnitude of the innovations should guarantee that
!               the sum of the coresponding z-thickness innovations is
!               equal to the sea-surface height innovation.
!
!               If constraints on minimum allowed layer thickness
!               prevents a sufficient part of one layer's pressure
!               thickness to be transferred to the other to accomodate
!               the ssh innovation, then that layers innovation will be
!               as (negatively) large as allowed, and the procedure will
!               be repeated for the  adjacent layer.
! PARAMETERS:
!     Name         Type     Usage           Description
!   ---------    --------   -----    -------------------------------
!    den         real       input    density profile (consistent with hycom density) 
!                                           in kg/m^3
!    dp          real       input    prior layer thickness in pressure units
!                                          (in pascal=kg/(m*s2))
!    ddp         real       output   pressure innovations
!    dpmin       real       input    minimum allowed layer thickness in press units
!    dzin        real       input    ssh innovation in meters
!    dzout       real       output   output (applyable) ssh innovation
!    g           real       input    gravitational constant
!    kmn         integer    input    index of upper-most layer with non-zero
!                                    innovation
!    kmx         integer    input    index of upper-most layer with non-zero
!                                    innovation
!    n_lyr       integer    input    max number layers
!
!

implicit  none

integer(i4)   ::   n_lyr
real(r4)      ::   deltap
real(r4)      ::   den (n_lyr)
real(r4)      ::   ddp(n_lyr)
real(r4)      ::   dp (n_lyr)
real(r4)      ::   dpmin(n_lyr)
real(r4)      ::   dz
real(r4)      ::   g
real(r4)      ::   thref, onem,dzin,dzout,sum1,sum2,eps,mindp
integer(i4)   ::   k, k1, k2, thisk1, thisk2, kmn, kmx, err, m
logical(bl)   ::   constraint

! check that no layer is thinner than dpmin, and kmn < kmx
      if(kmn.ge.kmx)then
!      write(6,*)'kmn.ge.kmx',kmn,kmx
      return
      endif
      do k=1,n_lyr
      if(dp(k).lt.dpmin(k))then
!      write(6,*)'dp < dpmin not allowed'
      return
      endif
      enddo
! acceptable error in sum dp
      eps=1.0  !mks pressure units ~ 1.e-4 meters
! initialize error to zero
      err=0
! layer-thickness innovations will be added to the following array
      do k = 1, n_lyr
         ddp(k) = 0.
      enddo

! put in dz the remaining innovation
      dz=dzin
      dzout=0.

! start iterations

! indices of upper-most and lower-most layers
      k1=kmn
      k2=kmx
!      write(6,*)'input k1,k2', k1,k2
      m=0
      do
      m=m+1

! k1 or k2 will be changed for next pass through do loop
      thisk1=k1
      thisk2=k2

! compute p-thickness adjustment to thisk1 (negative of that for p2)
! that will give z-thicknesses that account for ssh innovation dz

      if(den(thisk2)-den(thisk1)==0.0) return

      deltap=dz*den(thisk1)*den(thisk2)*g/(den(thisk2)-den(thisk1))
!      write(6,*)'m,thisk1,thisk2,deltap',m,thisk1,thisk2,deltap

! check the minimum thickness constraints

      constraint=.false.

      if (deltap .lt. dpmin(thisk1)-dp(thisk1))then
!      write(6,*)'thisk1,dpmin,dp,dpmin-dp',dpmin(thisk1),dp(thisk1),
!     .           dpmin(thisk1)-dp(thisk1)
! upper-most layer cannot be made thin enough
       constraint=.true.
! make it as thin as possible
       deltap=dpmin(thisk1)-dp(thisk1)
!        write(6,*)'constrained deltap',deltap
! increment upper-most layer for next iteration
       k1=k1+1
!       if(k1.ge.kmx) exit !BUG 
        if(k1.ge.thisk2) exit
      endif
!       write(6,*)'new k1',k1

      if (-deltap .lt. dpmin(thisk2)-dp(thisk2)) then
!        write(6,*)'thisk2,dpmin-dp',dpmin(thisk2)-dp(thisk2)
! lower-most layer cannot be made thin enough
       constraint=.true.
! make it as thin as possible
       deltap = -(dpmin(thisk2) - dp(thisk2))
!        write(6,*)'constrained deltap',deltap

! decrement lower-most layer for next iteration
       k2=k2-1
!      if(k2.le.kmn)exit !BUG
       if(k2.le.thisk1) exit
      endif
!      write(6,*)'new k2',k2


!     write(6,*)'after constraining', deltap

! add to innovation vector
      ddp(thisk1)=ddp(thisk1)+deltap
      ddp(thisk2)=ddp(thisk2)-deltap
!      write(6,*)'deltap',deltap
!      write(6,*)'ddp after this correction'
!      write(6,*)thisk1,ddp(thisk1),ddp(thisk1)/(den(thisk1)*g)
!      write(6,*)thisk2,ddp(thisk2),ddp(thisk2)/(den(thisk2)*g)
! ssh correction. 
!      write(6,*) 'ssh correction',
!     . deltap/(den(thisk1)*g)-deltap/(den(thisk2)*g)

! if constraints were not encountered, stop
      if(constraint.eqv..false.) exit

! if constraints were encountered, add applyable innovation,
! replace ssh innovation with the part that has not be accounted for
      dzout=dzout+deltap*(den(thisk2)-den(thisk1))/(g*den(thisk1)*den(thisk2))
!     write(6,*)'dz',dz
      dz=dz-deltap*(den(thisk2)-den(thisk1))/(g*den(thisk1)*den(thisk2))
!      write(6,*)'remaining dz after present correction',dz

      enddo
! add applyable innovation at the end of the subroutine
      dzout=dzout+deltap*(den(thisk2)-den(thisk1))/(g*den(thisk1)*den(thisk2))
! remaining unaccounted dz at the end of the subroutine
      dz=dz-deltap*(den(thisk2)-den(thisk1))/(g*den(thisk1)*den(thisk2))
!     write(6,*)'remaining dz at end of subroutine',dz
      sum1=0
      sum2=0
      mindp=0.
      do k=kmn,kmx
      if(dp(k)+ddp(k)-dpmin(k).lt.-eps.and.ddp(k).ne.0.)then
!      write(6,*)'error,k,dp,ddp,dp+ddp,dpmin'
!      write(6,*)k,dp(k),ddp(k),dp(k)+ddp(k),dpmin(k)
      if(dp(k)+ddp(k).lt.0) write(6,*)'fatal error, negative output dp',dp(k)+ddp(k)
      
      err=1
      endif
      sum1=sum1+ddp(k)
      sum2=sum2+ddp(k)/(g*den(k))
      enddo
!     write (6,*) sum1, sum2 
!     do k=1,n_lyr
!     write(6,*)k,dp(k)/9806.,ddp(k)/9806.
!     enddo
      if(abs(sum1).gt.eps)then
      err=1
!      write(6,*)'correction sum ddp (should be 0) and sum heights (ssh)', sum1,sum2
!      write(6,*)'output: dpp, dpp/(g*den)'
!      do k=kmn,kmx
!      write(6,*) i,j,k,ddp(k),ddp(k)/(g*den(k))
!      enddo
!      stop
      endif

! now ddp should contain pressure-thickness innovations for all layers

return
end subroutine

SUBROUTINE hybrid_min_thickness(nhybrd,nsigma,dp00,dp00x,dp00f,ds00,ds00x,ds00f,dp0k,ds0k)
IMPLICIT NONE

! Input hybrid grid related variables
INTEGER(i4)                   ::      nsigma,nhybrd
REAL(r4)                      ::      dp00,dp00x,dp00f,ds00,ds00x,ds00f
REAL(r4)                      ::      dp0kf,dpms,ds0kf,dsms,dpck(nhybrd)
REAL(r4)                      ::      dp0k(nhybrd),ds0k(nsigma)
REAL(r4)                      ::      sigscl(nhybrd),dsck(nhybrd),sigcum(nhybrd) 
REAL(r4)                      ::      depth,dmin
REAL(r4)                      ::      sigma(nhybrd)
LOGICAL(bl)                   ::      isopyc
INTEGER(i4)                   ::      k,nsigma_in

!nsigma_in=nsigma

      IF (nsigma.le.1) THEN

        nsigma=1
        ds00  =dp00
        ds00x =dp00x
        ds00f =dp00f

      ENDIF
 

!      IF(isopyc) THEN
!        dp0k(1)=thkmin
!      ELSE
        dp0k(1)=dp00
!      ENDIF

        dpms=dp0k(1)
        dp0kf=1.0
!        dpck(0)=0.0
        dpck(1)=dpms

      DO K=2,nhybrd
        dp0kf=dp0kf*dp00f
        IF(K.LE.nhybrd) THEN
          dp0k(K)=MIN(dp00*dp0kf,dp00x)
        ELSE
          dp0k(k)=0.0
        ENDIF
          dpms=dpms+dp0k(K)
          dpck(K)=dpms
        ENDDO

!     MINIMUM (SHALLOW) LAYER THICKNESSES.

!      IF (isopyc) THEN
!        ds0k(1)=thkmin
!      ELSE
        ds0k(1)=ds00
!      ENDIF

      dsms=ds0k(1)
      ds0kf=1.0
!      dsck(0)=0.0
      dsck(1)=dsms

      DO k=2,nsigma
        ds0kf=ds0kf*ds00f
        ds0k(K)=MIN(ds00*ds0kf,ds00x)
        dsms=dsms+ds0k(K)
        dsck(k)=dsms
      ENDDO

!      DO k=1,nsigma
!        sigscl(K)=ds0k(k)/dsms
!        sigcum(K)=dsck(k)/dsms
!      ENDDO

!      DO k= nsigma+1,nl
!        ds0k(k)=dp0k(k)
!        sigscl(k)=0.0
!        sigcum(k)=1.0
!      ENDDO

       !nsigma=nsigma_in
END SUBROUTINE

subroutine geocorr(start,count,mlat,mlon,dmgp,athk,auvl,avvl)
use mncio

implicit none

INTEGER(i4)         :: start(3), count(3),fid

integer(i4) :: i,j,k,nx,ny,nz
real(r4)    :: pi,f,f0,unit,finv,latmin,latmax,scvy1,scvy2,scux1,scux2

real(r4)    :: mlon(count(1),count(2))
real(r4)    :: mlat(count(1),count(2))
real(r4)    :: dmgp(count(1),count(2),count(3))
real(r4)    :: athk(count(1),count(2),count(3))
real(r4)    :: auvl(count(1),count(2),count(3))
real(r4)    :: avvl(count(1),count(2),count(3))

real(r4)    :: delu(count(1),count(2),count(3))
real(r4)    :: delv(count(1),count(2),count(3))
real(r4)    :: dptot(count(1),count(2))
real(r4)    :: worku(count(1),count(2))
real(r4)    :: workv(count(1),count(2))
logical(bl) :: land
REAL(r4)    ::  mscx(count(1),count(2)) ! pscx
REAL(r4)    ::  mscy(count(1),count(2)) ! pscy
real(r4),parameter :: flag=2.00**100
character(:),allocatable  :: gfile

nx=count(1)
ny=count(2)
nz=count(3)

gfile="gridinfo.nc"

! read grid file for pscx and pscy
call  nciopn(gfile,fid)
call  nciorv(gfile,fid,"mpscx",mscx,(/start(1),start(2)/),(/count(1),count(2)/))
call  nciorv(gfile,fid,"mpscy",mscy,(/start(1),start(2)/),(/count(1),count(2)/))
CALL  nciocl(gfile,fid)
latmin = 3.
latmax = 6.
pi     = 2. * asin(1.0)
f0     = 4. * pi / 24. / 3600.

dptot(1:nx,1:ny)=0.0
do k=1,nz
  dptot(1:nx,1:ny) = dptot(1:nx,1:ny) +athk(1:nx,1:ny,k)
enddo

   delu(:,:,:)=0.
   delv(:,:,:)=0.

   do i=2,nx-1
     do j=2,ny-1
! no corrections within 3 deg of the equator

       f = f0 * sin (mlat(i,j) * pi / 180. )
       if (abs(mlat(i,j)).gt.latmax) then
         finv = 1.0 / f
       elseif (abs(mlat(i,j)).lt.latmin) then
         finv = 0.0
       else
         finv = 1.0 / f * (abs(mlat(i,j))-latmin)/(latmax-latmin)
       endif

! calculate delu and delv at q grid points

       if (mscy(i-1,j).gt.0.) then
         scvy1=finv/(mscy(i-1,j))
       else
         scvy1=0.
       endif
       if (mscy(i,j).gt.0.) then
         scvy2=finv/(mscy(i,j))
       else
         scvy2=0.
       endif

       if (mscx(i,j-1).gt.0.) then
         scux1=finv/(mscx(i,j-1))
       else
         scux1=0.
       endif
       if (mscx(i,j).gt.0.) then
         scux2=finv/(mscx(i,j))
       else
         scux2=0.
       endif

       land = dptot(i-1,j-1).EQ.0.0 .or. dptot(i-1,j).EQ.0.0 .or. &
              dptot(i  ,j-1).EQ.0.0 .or. dptot(i  ,j).EQ.0.0

       if (.not.land) then
         do k=1,nz
           delu(i,j,k)= -0.5* &
                        ((dmgp(i-1,j  ,k)-dmgp(i-1,j-1,k))*scvy1 &
                        +(dmgp(i  ,j  ,k)-dmgp(i  ,j-1,k))*scvy2)
!if(i==420.and.j==170) then
!print *,'delu',k,delu(i,j,k),dmgp(i-1,j-1,k),dmgp(i,j-1,k),dmgp(i-1,j,k),dmgp(i,j,k)
!print *,montglf(i-1,j-1,k),xf%montgl(i-1,j-1,k,1)
!endif
           delv(i,j,k)=  0.5* &
                        ((dmgp(i  ,j-1,k)-dmgp(i-1,j-1,k))*scux1 &
                        +(dmgp(i  ,j  ,k)-dmgp(i-1,j  ,k))*scux2)
         enddo
       endif

     enddo
   enddo

! calculate delu on u grid points and delv on v grid points

   do k=1,nz

    worku(:,:) = 0.0
    workv(:,:) = 0.0

     do j=2,ny-1
       do i=2,nx-1
           worku(i,j) = 0.5*(delu(i,j,k)+delu(i,j+1,k))
           workv(i,j) = 0.5*(delv(i,j,k)+delv(i+1,j,k))
       enddo
     enddo

     delu(:,:,k)=worku(:,:)
     delv(:,:,k)=workv(:,:)

   enddo

    where(abs(delu)>1.0) delu=1.0
    where(abs(delv)>1.0) delv=1.0

    auvl=delu+auvl
    avvl=delv+avvl


  print *,'performed geostrophic velocity adjustment'

end subroutine


SUBROUTINE writearchv(archvFile,nx,ny,nz,time,thbase,thflag,sver,mlayer,mdepth,ssh,mld,uba,vba,uvl,vvl,thk,tem,sal)
USE mkinds
USE mabio
USE mdates
IMPLICIT NONE

! Argument Declarations

CHARACTER(len=*), INTENT(IN)           :: archvFile
INTEGER(i4),      INTENT(IN)           :: nx
INTEGER(i4),      INTENT(IN)           :: ny
INTEGER(i4),      INTENT(IN)           :: nz
REAL(r4)                               :: thbase
INTEGER(i4),      INTENT(IN)           :: sver
REAL(r4),         INTENT(INOUT)        :: mlayer(:)
REAL(r4),         INTENT(INOUT)        :: mdepth(:,:)
REAL(r4),         INTENT(INOUT)        :: ssh(:,:)
REAL(r4),         INTENT(INOUT)        :: mld(:,:)
REAL(r4),         INTENT(INOUT)        :: uba(:,:)
REAL(r4),         INTENT(INOUT)        :: vba(:,:)
REAL(r4),         INTENT(INOUT)        :: uvl(:,:,:)
REAL(r4),         INTENT(INOUT)        :: vvl(:,:,:)
REAL(r4),         INTENT(INOUT)        :: thk(:,:,:)
REAL(r4),         INTENT(INOUT)        :: tem(:,:,:)
REAL(r4),         INTENT(INOUT)        :: sal(:,:,:)

! Local Variables

INTEGER(i4)                            :: iunit,iyy,idd,imm,idoy,ihh
INTEGER(i4)                            :: k,ii,jj,kk
REAL(r4)                               :: hmin
REAL(r4)                               :: hmax
REAL(r4),PARAMETER                     :: spval=2.00**100
REAL(r4),PARAMETER                     :: hspval=0.5*spval
REAL(r4),PARAMETER                     :: qonem=1/9806.
REAL(r4)                               :: uzero
REAL(r4)                               :: vzero
REAL(r4)                               :: dummy(nx,ny)
REAL(r4)                               :: u_btrop(nx,ny)
REAL(r4)                               :: v_btrop(nx,ny)
CHARACTER(len=80), DIMENSION(5)        :: preambl
CHARACTER(len=80)                      :: cline
character(len=80)                      :: ctitle(4)
integer(i4)                            :: yrflag,iversn,sigver,iexpt,nstep,thflag
real(r4)                               :: time,coord



! Executable Section

iunit=11


! calculate thick weighted sums and remove non-zero barotropic velocities 

!do jj=1,ny
! do ii=1,nx

!    if(mdepth(ii,jj) .lt. hspval) then

!     uzero  = 0.0
!     vzero  = 0.0

!     do k= 1,nz
!     uzero = uzero + uvl(ii,jj,k)*thk(ii,jj,k)*qonem
!     vzero = vzero + vvl(ii,jj,k)*thk(ii,jj,k)*qonem
!     enddo

!     if(abs(uzero) > 0.0 .and. mdepth(ii,jj) < hspval) then
!     uzero = uzero / mdepth(ii,jj)
!     endif

!     if(abs(vzero) > 0.0 .and. mdepth(ii,jj) < hspval) then
!     vzero = vzero / mdepth(ii,jj)
!     endif


!     uba(ii,jj)=uzero
!     vba(ii,jj)=vzero

!     do k= 1,nz
!     uvl(ii,jj,k) = uvl(ii,jj,k) - uzero
!     vvl(ii,jj,k) = vvl(ii,jj,k) - vzero
!     enddo !k

!     else

!     uba(ii,jj)=0.0!spval
!     vba(ii,jj)=0.0!spval

!     do k= 1,nz
!     uvl(ii,jj,k) = 0.0!spval
!     vvl(ii,jj,k) = 0.0!spval
!     enddo !k
!     endif


! enddo
!enddo

!do kk=1,nz
! do jj=1,ny
!  do ii=2,nx
!
!     if(mdepth(ii-1,jj) < hspval  .and. mdepth(ii,jj) < hspval) then
!        uvl(ii,jj,kk)=0.5*(uvl(ii-1,jj,kk)+uvl(ii,jj,kk))
!     endif

!  enddo
!enddo
!enddo


!do kk=1,nz
! do jj=2,ny
!  do ii=1,nx

!     if(mdepth(ii,jj-1) < hspval  .and. mdepth(ii,jj) < spval) then
!        vvl(ii,jj,kk)=0.5*(vvl(ii,jj-1,kk)+vvl(ii,jj,kk))
!     endif

!  enddo
! enddo
!enddo

! ubaro

!u_btrop=uba

! do jj=1,ny
!  do ii=2,nx

!     if(mdepth(ii-1,jj) < hspval  .and. mdepth(ii,jj) < spval) then
!       u_btrop(ii,jj)=0.5*(uba(ii-1,jj)+uba(ii,jj))
!     else
!      u_btrop(ii,jj)=0.0!spval
!     endif
!  enddo
! enddo


! vbaro

! v_btrop=vba

! do jj=2,ny
!  do ii=1,nx

!  if(mdepth(ii,jj-1) < spval  .and. mdepth(ii,jj) < spval) then
!       v_btrop(ii,jj)=0.5*(vba(ii,jj-1)+vba(ii,jj))
!     else
!      v_btrop(ii,jj)=0.0!spval
!     endif
!  enddo
! enddo

!where(mdepth < spval) 
! dummy=0.0
!elsewhere
! dummy = 0.0
!endwhere

!where(abs(u_btrop) > 1.e3) u_btrop=0.0
!where(abs(v_btrop) > 1.e3) v_btrop=0.0

!where(abs(uvl) > 1.e3) uvl=0.0
!where(abs(vvl) > 1.e3) vvl=0.0

! write out some info for the bfile

ctitle(1)="TSIS Analysis for incremental update"
ctitle(2)="GDEM3 Jan. init; KPP; SeaWiFS KPAR; e-loan ice; A=20; Smag=.05; veldf4=.02;"
ctitle(3)="S-Z(14-18):dp00/f/x/i=3m/1.18/120m/1m;ds=0.5m/1.18/75m; var den: Med+Black Seas;"
ctitle(4)="Dummy Info in the above two lines"
iversn=00
iexpt=000
yrflag=3

call zaiost(nx,ny)
call zaiopn(trim(archvFile),iunit)
!!Alex write(iunit,116) ctitle,iversn,iexpt,iexpt,yrflag,nx,ny
write(iunit,116) ctitle,iversn,iexpt,yrflag,nx,ny
print *,'writing incupd files ',ctitle(1)

call zaiowr(iunit,ssh,hmin,hmax)
WRITE (iunit,117) 'montg1  ',nstep,time,sver,thbase,hmin,hmax

call zaiowr(iunit,ssh,hmin,hmax)
WRITE (iunit,117) 'srfhgt  ',nstep,time,0,0,hmin,hmax

!call zaiowr(iunit,dummy,hmin,hmax)
!WRITE (iunit,117) 'steric  ',nstep,time,0,0,hmin,hmax

call zaiowr(iunit,ssh,hmin,hmax)
WRITE (iunit,117) 'surflx  ',nstep,time,0,0,hmin,hmax

call zaiowr(iunit,ssh,hmin,hmax)
WRITE (iunit,117) 'salflx  ',nstep,time,0,0,hmin,hmax

call zaiowr(iunit,ssh,hmin,hmax)
WRITE (iunit,117) 'bl_dpth  ',nstep,time,0,0,hmin,hmax

call zaiowr(iunit,ssh,hmin,hmax)
WRITE (iunit,117) 'mix_dpth  ',nstep,time,0,0,hmin,hmax

call zaiowr(iunit,uba,hmin,hmax)
WRITE (iunit,117) 'u_btrop  ',nstep,time,0,0,hmin,hmax

call zaiowr(iunit,vba,hmin,hmax)
WRITE (iunit,117) 'v_btrop  ',nstep,time,0,0,hmin,hmax

!     layer loop

DO k=1,nz

coord=mlayer(k)

CALL zaiowr(iunit,uvl(:,:,k),hmin,hmax)
WRITE (iunit,117) 'u-vel.  ',nstep,time,k,coord,hmin,hmax

CALL zaiowr(iunit,vvl(:,:,k),hmin,hmax)
WRITE (iunit,117) 'v-vel.  ',nstep,time,k,coord,hmin,hmax

CALL zaiowr(iunit,thk(:,:,k),hmin,hmax)
WRITE (iunit,117) 'thknss   ',nstep,time,k,coord,hmin,hmax

CALL zaiowr(iunit,tem(:,:,k),hmin,hmax)
WRITE (iunit,117) 'temp    ',nstep,time,k,coord,hmin,hmax

CALL zaiowr(iunit,sal(:,:,k),hmin,hmax)
WRITE (iunit,117) 'salin   ',nstep,time,k,coord,hmin,hmax

ENDDO

call zaiocl(iunit)

116  format (a80/a80/a80/a80/ &
i5,4x,'''iversn'' = hycom version number x10'/ &
i5,4x,'''iexpt '' = experiment number x10'/    &
!!Alex i5,4x,'''jexpt '' = experiment number x10'/    &
i5,4x,'''yrflag'' = days in year flag'/        &
i5,4x,'''idm   '' = longitudinal array size'/  &
i5,4x,'''jdm   '' = latitudinal  array size'/  &
!!Alex'field       time step   diff day','  k  dens        min              max')
'field       time step  model day','  k  dens        min              max')

117  FORMAT (a8,' =',i11,f11.3,i3,f6.2,1p2e16.7)

END SUBROUTINE

subroutine layerizei(kz,zlevin,tzin,szin,rzin,uzin,vzin,depth,nl,pl,rl,sl,  &
                    tl,ul,vl,thbase,thkmin,nhybrd,nsigma,dp00, &
                    dp00x,dp00f,ds00,ds00x,ds00f,dpk,dsk,ilspec,sigma,isotop,sigver)

USE mkinds
use mseawater

IMPLICIT NONE

! Input hybrid grid related variables

INTEGER(i4)                   ::      nhybrd,nsigma,nl,thflg,sigver
REAL(r4)                      ::      dp00s,dp00,dp00x,dp00f,ds00,ds00x,ds00f,dp00i,thkmin
REAL(r4)                      ::      dp0kf,dpms,ds0kf,dsms,dpck(0:nl),dsck(0:nl),sigcum(0:nl)
REAL(r4),intent(in),optional  ::      dpk(nhybrd),dsk(nsigma)
!REAL(r4)                      ::      dp0k(0:nl),ds0k(0:nl),sigscl(0:nl),dsck(0:nl),sigcum(0:nl)
REAL(r4)                      ::      dp0k(nl),ds0k(nl),sigscl(nl)
REAL(r4)                      ::      depth,dmin,isotop,thbase
REAL(r4)                      ::      sigma(nl)
LOGICAL                       ::      isopyc
integer(i4)                   ::      ilspec
! Output

REAL(r4)                      ::      pmix,pm,tm,sm,rm,um,vm

! local variables

INTEGER(i4)                   ::      msk

REAL(r4)                      ::      pkm1,pkm2,rkm1,rkm2,work
REAL(r4)                      ::      szk,rzk,tzk,time,thk
INTEGER(i4)                   ::      i,j,k,k400,kz,kzl,kk,l
REAL(r4)                      ::      pave,xave,xmin,xmax,tmin,smin,pmin,zj,zz,q
REAL(r4)                      ::      pinteg,sigmaa,sigmab,zbot,ztop,sigu,sigv
REAL(r4)                      ::      zlev(kz+1),zlevin(kz)
REAL(r4)                      ::      tzin(kz),rzin(kz),szin(kz),uzin(kz),vzin(kz)
REAL(r4)                      ::      tz(kz+1),rz(kz+1),sz(kz+1),uz(kz+1),vz(kz+1)
REAL(r4)                      ::      tl(nl),rl(nl),sl(nl),pl(nl+1),vl(nl),ul(nl)

real(r4),parameter            ::      zero=0.0,one=1.0

real(r4)                      ::      qdep,pztop,pzbot,rzloc,pzmid,pcm0,pcm1,pcm2
integer(r4)                   ::      kztop

zlev(1:kz)=zlevin
tz(1:kz)=tzin
sz(1:kz)=szin
rz(1:kz)=rzin
uz(1:kz)=uzin
vz(1:kz)=vzin



if (ilspec==1) then

dp0K=dpk
dpms=dp0K(1)
DO K=2,nl
   dpms=dpms+dp0k(K)
   dpck(K)=dpms
enddo

ds0k=dsk
dsms=ds0k(1)
dsck(1)=dsms

 DO k=2,nsigma
     dsms=dsms+ds0k(K)
     dsck(k)=dsms
 ENDDO

 DO k=1,nsigma
    sigscl(K)=ds0k(k)/dsms
    sigcum(K)=dsck(k)/dsms
 ENDDO

 DO k= nsigma+1,nl
    ds0k(k)=dp0k(k)
    sigscl(k)=0.0
    sigcum(k)=1.0
 ENDDO

else
   IF (nsigma.le.1) THEN

        nsigma=1
        ds00  =dp00
        ds00x =dp00x
        ds00f =dp00f

      ENDIF
IF (nsigma.le.1) THEN

        nsigma=1
        ds00  =dp00
        ds00x =dp00x
        ds00f =dp00f

      ENDIF

     ! IF(isopyc) THEN
     !   dp0k(1)=thkmin
     ! ELSE
        dp0k(1)=dp00
     ! ENDIF

        dpms=dp0k(1)
        dpck(0)=0.0
        dpck(1)=dpms
        dp0kf=one


      DO K=2,nl
        dp0kf=dp0kf*dp00f
        IF(K.LE.nhybrd) THEN
          dp0k(K)=MIN(dp00*dp0kf,dp00x)
        ELSE
          dp0k(k)=zero
        ENDIF
          dpms=dpms+dp0k(K)
          dpck(K)=dpms
        ENDDO
!     MINIMUM (SHALLOW) LAYER THICKNESSES.

!      IF (isopyc) THEN
!        ds0k(1)=thkmin
!      ELSE
        ds0k(1)=ds00
!      ENDIF

       dsms=ds0k(1)
       dsck(0)=0.0
       dsck(1)=dsms
       ds0kf=one

      DO k=2,nsigma
        ds0kf=ds0kf*ds00f
        ds0k(K)=MIN(ds00*ds0kf,ds00x)
        dsms=dsms+ds0k(K)
        dsck(k)=dsms
      ENDDO
      sigcum(0)=0.0
      DO k=1,nsigma
        sigscl(K)=ds0k(k)/dsms
        sigcum(K)=dsck(k)/dsms
      ENDDO

      DO k= nsigma+1,nl
        ds0k(k)=dp0k(k)
        sigscl(k)=0.0
        sigcum(k)=1.0
      ENDDO

endif
do k=1,nl
WRITE(6,6000) k,DP0K(k),ds0k(k),SIGMA(k),' Sig2'
enddo

ISOPYC=.false.

      ZLEV(KZ+1) = ZLEV(KZ) + 10000.0

      DO k= 1,kz
      IF (rz(k).GT. sigma(nl)) THEN

         IF (rz(max(k-1,1)).EQ. sigma(nl)) THEN
                  rz(k) = sigma(nl)
                  tz(K) = tz(k-1)
                ELSE
                  sz(k) = sofsig(rz(k),tz(k),sigver)
                  rz(k) = sigma(nl)
                  tz(K) = tofsig(sigma(nl),sz(k),sigver)
                ENDIF
              ENDIF
      ENDDO

          RZ(KZ+1) = RZ(KZ) + 0.001
          TZ(KZ+1) = TZ(KZ)
          SZ(KZ+1) = SZ(KZ)
          UZ(KZ+1) = UZ(KZ)
          VZ(KZ+1) = UZ(KZ)


!
     IF (depth.GT.zero) THEN
        rkm1 = zero
        pkm1 = zero
        pkm2 = zero
        PCM0 = ZERO
        PCM1 = ZERO
        PCM2 = ZERO
        PZTOP = 1.0  !surface
        PZBOT = 1.0  !surface
     ENDIF

  DO 810 K= 1,nl


      IF(depth.GT.zero) THEN         ! depth > loop --line 206



   IF(k.GT.1) THEN
    rm = sigma(k)
   ELSE
    rm = min(sigma(1),rz(1))
   ENDIF

   QDEP = MAX(0.0,MIN(1.0,(depth-DSCK(NSIGMA))/(DPCK(NSIGMA)-DSCK(NSIGMA))))
   DMIN = (1.0-QDEP)*DSCK(K-1) + QDEP*DPCK(K-1)

 
   IF(isopyc .AND. k.EQ.1) THEN
!  UPPER LAYER IS THE MIXED LAYER.
   
   PM = MIN(DEPTH,PMIX)

   ELSEIF (k.EQ.nl) THEN
!  lowest layer

   pm = depth

   ELSEIF (dmin.le.isotop) THEN

!FIXED GRID NEAR THE SURFACE.

    DMIN=(1.0-QDEP)*DSCK(K) + QDEP*DPCK(K)
    PM=MIN(DEPTH, MAX(PKM1,DMIN))

!  WRITE(6,'(A,I3,F10.3,F10.3,F10.3)') 'FIXED - K,PM =',K,PM,PKM1,rm

  ELSEIF (rm.LT.rkm1) THEN
! ZERO THICKNESS LAYER REQUIRED FOR STABILITY.

   pm = pkm1

   ELSE  !K.LT.nl

! INTERFACE IS HALF WAY BETWEEN TARGET DENSITIES.

   SIGMAA = 0.5*(sigma(k) + sigma(k+1))
   KZTOP  = PZTOP
  

   CALL FIND_DENSITY(PZBOT,SIGMAA,RZ,KZ+1,KZTOP)
  

   IF(PZBOT.EQ.0.0) THEN
      PM = PKM1
   ELSEIF (PZBOT.EQ.KZ+1) THEN
      PM = DEPTH
   ELSE
      L = PZBOT
      Q = PZBOT - L
      PM = MIN(DEPTH,MAX(PKM1,(1.0-Q)*zlev(L) + Q*zlev(L+1)))
   ENDIF
   
   ENDIF ! exactly isopycnal

!  MODIFY RM AND PM (IF NECESSARY).
   
   KZTOP=PZTOP
   CALL FIND_DEPTH(PZBOT,PM,ZLEV,KZ+1,KZTOP)
   SIGMAA = sigma(K)
   KZTOP  = PZTOP
   CALL FIND_DENSITY(RZLOC,SIGMAA,RZ,KZ+1,KZTOP)
   IF(RZLOC.LT.PZTOP .OR. RZLOC.GT.PZBOT) THEN
      PZMID = 0.5*(PZTOP + PZBOT)
   ELSE
      PZMID = RZLOC
   ENDIF

  IF(DMIN.LE.ISOTOP) THEN
         DPMS=PM
         PCM0=DPMS
  ELSE
    
      IF(DP0K(K).LE.DP00I) THEN
          Q = DP0K(K)
      ELSE
          Q  = MAX(DP00I,DP0K(K)*DP0K(K)/MAX( DP0K(K),PKM1-PCM1))
      ENDIF
    PCM0=PCM1 + MIN(Q,(1.0-QDEP)*DS0K(K)+QDEP*DP0K(K))
    DPMS=PKM1 + MIN(Q,(1.0-QDEP)*DS0K(K)+QDEP*DP0K(K))
  ENDIF

    IF(.NOT. ISOPYC .AND. PM .EQ. DEPTH) THEN
        
     IF(PM.GT.PKM1) THEN
!
!MAKE LOWEST NON-ZERO LAYER EXACTLY ISOPYNCAL,UNLESS THE LAYER ABOVE IS NOT ISOPYCNAL.
!
       IF(K.GT.1) THEN
           IF(RKM1 .NE.sigma(K-1)) THEN
             THK=PKM1-PKM2
              IF(DP0K(K-1).LE.DP00I) THEN
                 Q = DP0K(K-1)
              ELSE
                 Q = MAX( DP00I,DP0K(K-1)*DP0K(K-1)/MAX(DP0K(K-1),PKM2-PCM2))
              ENDIF
              THKMIN = MIN(Q,(1.0-QDEP)*DS0K(K)+QDEP*DP0K(K))
            ELSE  ! LAYER ABOVE IS DEFINITELY ISOPYCNAL
              THK    = 1.0
              THKMIN = 0.0
           ENDIF
         ELSE
               THK    = 1.0
               THKMIN = 1.0
        ENDIF

        IF(THKMIN .EQ.ZERO .OR. ABS(THK-THKMIN).GT.0.01) THEN
              RM = MAX(RKM1,sigma(k))
        ELSE
        L = PZMID
        Q = PZMID - L
        RM = (1.0-Q)*RZ(L) + Q*RZ(L+1)
        ENDIF
                
       ELSE
        RM = RKM1
       ENDIF

   ELSEIF (.NOT. ISOPYC .AND. PM .LE. MIN(DEPTH,DPMS)) THEN

!               HYBRID (MINIMUM THICKNESS) LAYER.

                PM =MIN(DPMS,DEPTH)
                KZTOP  = PZTOP
                CALL FIND_DEPTH(PZBOT,PM,Zlev,KZ+1,KZTOP)
                IF(RZLOC.LT.PZTOP .OR. RZLOC.GT.PZBOT) THEN
                  PZMID = 0.5*(PZTOP + PZBOT)
                ELSE
                  PZMID = RZLOC
                ENDIF
                IF(PM.GT.PKM1) THEN
                  L = PZMID
                  Q = PZMID - L
                  RM = (1.0-Q)*RZ(L) + Q*RZ(L+1)
                ENDIF
              ELSE

!               ISOPYCNAL LAYER, BUT RECALCULATE DENSITY ANYWAY.
!
                IF (PM .GT.PKM1) THEN
                  L = PZMID
                  Q = PZMID - L
                  RM=(1.0-Q)*RZ(L) + Q*RZ(L+1)
                ENDIF
              
        ENDIF
 
!             FIND T & S.

              IF(PM .GT. PKM1) THEN
                L = PZMID
                Q = PZMID - L
                TM = (1.0-Q)*TZ(L) + Q*TZ(L+1)
                UM = (1.0-Q)*UZ(L) + Q*UZ(L+1)
                VM = (1.0-Q)*VZ(L) + Q*VZ(L+1)
              ENDIF
             
              sm = sofsig(rm,tm,sigver)
        endif
! PREPARE FOR NEXT K-LOOP

     pkm2 = pkm1
     pkm1 = pm
     rkm1 = rm
     PCM2 = PCM1
     PCM1 = PCM0
     PZTOP = PZBOT
     PZBOT = 0.0

    pl(k+1)=pm
    tl(k)=tm
    sl(k)=sm
    rl(k)=rm
    ul(k)=um
    vl(k)=vm
 810 CONTINUE


 6000 FORMAT('k =',I3,'   thkns =',F6.1,' m','   depth =',F7.1,' m','   density =',F7.3,A)




end subroutine


SUBROUTINE layerize(kz,zlev,tz,sz,rz,uz,vz,depth,nl,pl,rl,sl,  &
                    tl,ul,vl,thbase,thkmin,nhybrd,nsigma,dp00, &
                    dp00x,dp00f,ds00,ds00x,ds00f,dpk,dsk,ilspec,sigma,isotop,sigver)

USE mkinds
use mseawater

IMPLICIT NONE

! Input hybrid grid related variables

INTEGER(i4)                   ::      nhybrd,nsigma,nl,thflg,sigver
REAL(r4)                      ::      dp00s,dp00,dp00x,dp00f,ds00,ds00x,ds00f,dp00i,thkmin
REAL(r4)                      ::      dp0kf,dpms,ds0kf,dsms,dpck(0:nl),dsck(0:nl),sigcum(0:nl)
REAL(r4),intent(in),optional  ::      dpk(nhybrd),dsk(nsigma)
!REAL(r4)                      ::      dp0k(0:nl),ds0k(0:nl),sigscl(0:nl),dsck(0:nl),sigcum(0:nl)
REAL(r4)                      ::      dp0k(nl),ds0k(nl),sigscl(nl)
REAL(r4)                      ::      depth,dmin,isotop,thbase
REAL(r4)                      ::      sigma(nl)
LOGICAL                       ::      isopyc
integer(i4)                   ::      ilspec
! Output

REAL(r4)                      ::      pmix,pm,tm,sm,rm,um,vm

! local variables

INTEGER(i4)                   ::      msk

REAL(r4)                      ::      pkm1,pkm2,rkm1,rkm2,work
REAL(r4)                      ::      zi(0:kz),zil(0:kz)
REAL(r4)                      ::      szk,rzk,tzk,time,thk

INTEGER(i4)                   ::      i,j,k,k400,kz,kzl,kk

REAL(r4)                      ::      pave,xave,xmin,xmax,tmin,smin,pmin,zj,zz,q
REAL(r4)                      ::      pinteg,sigmaa,sigmab,zbot,ztop,sigu,sigv


REAL(r4)                      ::      zlev(kz+1)
REAL(r4)                      ::      tz(kz+1),rz(kz+1),sz(kz+1),uz(kz+1),vz(kz+1)
REAL(r4)                      ::      tl(nl),rl(nl),sl(nl),pl(nl+1),vl(nl),ul(nl)

real(r4),parameter            ::      zero=0.0,one=1.0

real(r4)                      ::      qdep



if (ilspec==1) then 

dp0K=dpk
dpms=dp0K(1)
DO K=2,nl
   dpms=dpms+dp0k(K)
   dpck(K)=dpms
enddo

ds0k=dsk
dsms=ds0k(1)
dsck(1)=dsms

 DO k=2,nsigma
     dsms=dsms+ds0k(K)
     dsck(k)=dsms
 ENDDO

 DO k=1,nsigma
    sigscl(K)=ds0k(k)/dsms
    sigcum(K)=dsck(k)/dsms
 ENDDO

 DO k= nsigma+1,nl
    ds0k(k)=dp0k(k)
    sigscl(k)=0.0
    sigcum(k)=1.0
 ENDDO

else
   IF (nsigma.le.1) THEN

        nsigma=1
        ds00  =dp00
        ds00x =dp00x
        ds00f =dp00f

      ENDIF

     ! IF(isopyc) THEN
     !   dp0k(1)=thkmin
     ! ELSE
        dp0k(1)=dp00
     ! ENDIF

        dpms=dp0k(1)
        dpck(0)=0.0
        dpck(1)=dpms
        dp0kf=one


      DO K=2,nl
        dp0kf=dp0kf*dp00f
        IF(K.LE.nhybrd) THEN
          dp0k(K)=MIN(dp00*dp0kf,dp00x)
        ELSE
          dp0k(k)=zero
        ENDIF
          dpms=dpms+dp0k(K)
          dpck(K)=dpms
        ENDDO
!     MINIMUM (SHALLOW) LAYER THICKNESSES.

!      IF (isopyc) THEN
!        ds0k(1)=thkmin
!      ELSE
        ds0k(1)=ds00
!      ENDIF
       dsms=ds0k(1)
       dsck(0)=0.0
       dsck(1)=dsms
       ds0kf=one

      DO k=2,nsigma
        ds0kf=ds0kf*ds00f
        ds0k(K)=MIN(ds00*ds0kf,ds00x)
        dsms=dsms+ds0k(K)
        dsck(k)=dsms
      ENDDO
      sigcum(0)=0.0
      DO k=1,nsigma
        sigscl(K)=ds0k(k)/dsms
        sigcum(K)=dsck(k)/dsms
      ENDDO

      DO k= nsigma+1,nl
        ds0k(k)=dp0k(k)
        sigscl(k)=0.0
        sigcum(k)=1.0
      ENDDO

endif

do k=1,nl
WRITE(6,6000) k,DP0K(k),ds0k(k),SIGMA(k),' Sig2'
enddo



!     CONVERT Z-LEVELS TO FINITE VOLUMES.

      zi(0) = 0.0!zero

      DO k= 1,kz-1
        zi(k) = 0.5*(zlev(K) + zlev(k+1))
      ENDDO

      zi(kz) = 9999.0

! find k400

!     DO K=1,kz
!     IF (K.GT.1) THEN
!       IF(zlev(k).GT.399.0 .AND. zlev(k-1).LT.399) THEN
!             k400 = k
!          ENDIF
!        ENDIF
!      ENDDO

!     LIMIT MAXIMUM DENSITY TO SIGMA(nl)

      DO k= 1,kz
      IF (rz(k).GT. sigma(nl)) THEN

         IF (rz(max(k-1,1)).EQ. sigma(nl)) THEN
                  rz(k) = sigma(nl)
                  tz(K) = tz(k-1)
                ELSE
                  sz(k) = sofsig(rz(k),tz(k),sigver)
                  rz(k) = sigma(nl)
                  tz(K) = tofsig(sigma(nl),sz(k),sigver)
                ENDIF
              ENDIF
      ENDDO
!
! IF(depth.GT.zero) THEN

!            pkm2 = zero
!            pkm1 = zero
!            tm   = tz(1)-0.15
!            sm   = sofsig(r8(rz(1)),R8(tz(1)))
!            sigmaa    = sig(r8(tz(1)-0.1),r8(sm))
!            pmix  = 400.0

!            DO k= 2,k400
!              IF     (rz(k).GE.sigmaa) THEN
!                sigmab = rz(K)-rz(k-1)
!                IF     (sigmab.GT.1.E-4) THEN
!                  q = (rz(k)-sigmaa)/sigmab
!                  pmix = zlev(k-1) + (1.0-q)*(zlev(k)-zlev(k-1))
!                ELSE
!                  pmix = zlev(k)
!                ENDIF
!                EXIT
!              ENDIF
!            ENDDO
!            pmix= MAX(thkmin, MIN(pmix,depth))
!          ELSE
!            pkm2 = 0.0
!            pkm1 = 0.0
!            pmix = thkmin
!            tm   = 0.0
!            sm   = 0.0
!            rm   = 0.0
!            pm   = 0.0
 !           depth= 0.0  ! should already be zero
 !         ENDIF
!
     IF (depth.GT.zero) THEN
        rkm1 = zero
        pkm1 = zero
        pkm2 = zero
     ENDIF

     DO 810 K= 1,nl

      IF(depth.GT.zero) THEN         ! depth > loop --line 206

!    ALLOW FOR BATHYMETRY.

         DO kzl= 0,kz
            zil(kzl) = zi(kzl)
             IF(zi(kzl).GT.depth) THEN  ! if  --line 212
                zil(kzl) = depth
                EXIT
             ENDIF                      ! endif  --215 
         ENDDO


    IF(k.EQ.1 .AND. kzl.GE.3) THEN
! MODIFY LOWEST FINITE VOLUME USING EXTRAPOLATION
! FROM ABOVE, BECAUSE LOWEST VALUES LIKELY TO BE
! BELOW THE BOTTOM AND THEREFORE INACCURATE.

    zz = 0.5*(zil(kzl-1)+zil(kzl))
    q  = (zz - zi(kzl-2))/(zi(kzl-1) - zi(kzl-2))
    rz(kzl) = (1.0-q)*rz(kzl-2) +  q *rz(kzl-1)
    tz(kzl) = (1.0-q)*tz(kzl-2) +  q *TZ(kzl-1)

    ENDIF

   IF(k.GT.1) THEN
    rm = sigma(k)
   ELSE
    rm = min(sigma(1),rz(1))
   ENDIF

   IF(nsigma.gt.1 .AND.depth .LT.dsck(nsigma)) THEN
   dmin = MIN(dsck(K-1),depth)
   ELSEIF (nsigma.GT.1 .AND.depth.LT.dpck(nsigma)) THEN
   dmin = sigcum(k-1)*depth
   ELSE
   dmin = MIN(dpck(k-1),depth)
   ENDIF


!!!
!!!
   QDEP = MAX(0.0,MIN(1.0,(depth-DSCK(NSIGMA))/(DPCK(NSIGMA)-DSCK(NSIGMA))))
   DMIN = (1.0-QDEP)*DSCK(K-1) + QDEP*DPCK(K-1)
!
!
!
!!!!!!!!!!


!   WRITE(6,'(A,I3,F10.3)') 'ISOTOP - K,DMIN =',K,DMIN

!   FIND RM AND PM.

   IF(isopyc .AND. k.EQ.1) THEN
!  UPPER LAYER IS THE MIXED LAYER.

!   sigmaa = zero
!   DO kk= 1,kzl
!   ztop   = zil(kk-1)
!   zbot   = max(ztop,min(zil(kk),pmix))
!   sigmaa = sigmaa + rz(kk)*(zbot-ztop)
!   ENDDO
!   rm = sigmaa/pmix
!   pm = pmix

!!!!!
   PM=MIN(DEPTH, PMIX)
!!!!!

   ELSEIF (k.EQ.nl) THEN
!  lowest layer

   pm = depth

   ELSEIF (dmin.le.isotop) THEN

!FIXED GRID NEAR THE SURFACE.

  IF(nsigma.GT.1 .AND.depth.LT.dsck(nsigma)) THEN
     pm = MAX(pkm1,MIN(dsck(k),depth) )
  ELSEIF (nsigma.GT.1 .AND.depth.LT.dpck(nsigma)) THEN
     pm = MAX(pkm1,sigcum(k)*depth)
  ELSE
     pm = MAX(pkm1,MIN(dpck(k),depth))
  ENDIF

!!!! 
!  DMIN=(1.0-QDEP)*DSCK(K) + QDEP*DPCK(K)
!  PM = MIN(DEPTH, MAX(PKM1,DMIN))
!!!!

!  WRITE(6,'(A,I3,F10.3,F10.3,F10.3)') 'FIXED - K,PM =',K,PM,PKM1,rm

  ELSEIF (rm.LT.rkm1) THEN
! ZERO THICKNESS LAYER REQUIRED FOR STABILITY.

   pm = pkm1

   ELSE  !K.LT.nl
! REMAP TO EXACTLY ISOPYCNAL LAYER
  pinteg  = zero
   sigmab  = rm

   DO kk= 1,kzl
   IF(zil(kk).GT.pkm1) THEN
   sigmaa = sigmab
   sigmab = min(sigma(k+1),max(rm,rz(kk)))
   pinteg = pinteg + (sigmab-sigmaa)*max(pkm1,zil(kk-1))
   ENDIF
   ENDDO

   sigmaa = sigmab
   sigmab = sigma(K+1)
   pinteg = pinteg + (sigmab-sigmaa)*max(pkm1,zil(kzl))

   IF(pinteg.GT.zero) THEN

   pm = MIN(depth,MAX(pkm1,pinteg/(sigma(k+1)-rm)))

   ELSEIF (rm.LT.rz(kzl)) THEN
!AT THE TOP.

   pm = zero

   ELSE

!AT THE BOTTOM.

   pm = depth

   ENDIF
















   ENDIF ! exactly isopycnal

   
   
   
   
!  MODIFY RM AND PM (IF NECESSARY).

   dpms = pkm1 + min(dp0k(K),max(ds0k(k),sigscl(k)*depth))

!  Lowest layer setting

   IF(.NOT. isopyc .AND. pm.EQ.depth) THEN

    IF(pm.GT.pkm1) THEN

!  MAKE LOWEST NON-ZERO LAYER EXACTLY ISOPYNCAL, UNLESS THE LAYER ABOVE IS NOT ISOPYCNAL.

   IF(k.GT.1) THEN

     IF(rkm1.NE.sigma(K-1)) THEN
      thk=pkm1-pkm2
      thkmin =min(dp0k(K-1),max(ds0k(k-1),sigscl(k-1)*depth))
     ELSE  ! LAYER ABOVE IS ISOPYCNAL
      thk    = 1.0
      thkmin = 0.0
     ENDIF

     ELSE

      thk    = 1.0
      thkmin = 1.0

   ENDIF

    IF(thkmin .EQ.ZERO .OR. abs(thk-thkmin).GT.0.01) THEN
      RM = max(rkm1,sigma(k))

    ELSE

!  LOWEST NON-ZERO LAYER IS NOT-ISOPYCNAL.

    sigmaa = zero
    sigmab = zero

    DO kk= 1,kzl
    ztop   = max(zil(kk-1),pkm1)
    zbot   = max(ztop,min(zil(kk),PM))
    sigmaa = sigmaa + rz(kk)*(zbot-ztop)
    ENDDO

    rm = max(rkm1,sigmaa/(pm-pkm1))
!    print *, "***",zbot,ztop,tz(1),tz(2)
    ENDIF

    ELSE

    rm = rkm1

    ENDIF

! lowest layer setting

    ELSEIF (.not. isopyc .and. pm.LE.min(depth,dpms)) THEN

!   HYBRID (MINIMUM THICKNESS) LAYER.

    pm = min(dpms,depth)
    sigmaa = zero
    DO kk= 1,kzl
    ztop= max(zil(kk-1),pkm1)
    zbot= max(ztop,min(zil(kk),pm))
    sigmaa = sigmaa + rz(kk)*(zbot-ztop)
    ENDDO

    rm=max(rkm1,sigmaa/(pm-pkm1))

!    WRITE(6,'(A,I3,3F10.3)') 'HYBRID - K,PKM1,PM,RM =',K,PKM1,PM,RM

    ELSE
!
! ISOPYCNAL LAYER, BUT RECALCULATE DENSITY ANYWAY.
!
   IF(pm.GT.pkm1) THEN
   sigmaa = zero
   DO kk= 1,kzl
   ztop   = MAX(zil(kk-1),pkm1)
   zbot   = MAX(ztop,MIN(zil(kk),pm))
   sigmaa = sigmaa + rz(kk)*(ZBOT-ZTOP)
   ENDDO
   rm = sigmaa/(pm-pkm1)
   ELSE
!  RM UNCHANGED
   ENDIF

   ENDIF

!   FIND T & S.

    IF(pm.GT.pkm1) THEN
    sigmaa = zero
    sigu=zero
    sigv=zero
 DO kk= 1,kzl
    ztop   = max(zil(kk-1),pkm1)
    zbot   = max(ztop,min(zil(kk),pm))
    sigmaa = sigmaa + tz(kk)*(zbot-ztop)
    sigu   = sigu   + uz(kk)*(zbot-ztop)
    sigv   = sigv   + vz(kk)*(zbot-ztop)
 ENDDO
    tm     = sigmaa/(pm-pkm1)
    um     = sigu/(pm-pkm1)
    vm     = sigv/(pm-pkm1)
 ENDIF
!   write(*,*) kzl
 !   write(*,*)

    sm = sofsig(rm,tm,sigver)
!    write(*,*) k,sm,tm,rm
 !   write(*,*)
 !   write(*,*)
 !   write(*,*) k,sm,tm,rm


 ENDIF  !DEPTH>0

! PREPARE FOR NEXT K-LOOP

     pkm2 = pkm1
     pkm1 = pm
     rkm1 = rm


    pl(k+1)=pm
    tl(k)=tm
    sl(k)=sm
    rl(k)=rm
    ul(k)=um
    vl(k)=vm
 810 CONTINUE


 6000 FORMAT('k =',I3,'   thkns =',F6.1,' m','   depth =',F7.1,' m','   density =',F7.3,A)
END SUBROUTINE

SUBROUTINE FIND_DENSITY(DENLOC,DENTARG,RZ,KZ,MINZ)
      IMPLICIT NONE
!
      INTEGER KZ,MINZ
      REAL*4  DENLOC,DENTARG,RZ(KZ)
!
!     FIND EXACT LOCATION IN LAYER SPACE OF DENTARG
!     SEARCHING WITHIN RZ(MINZ:KZ)
!
!     ASSUME RZ IS MONOTONICALLY NON-DECREASING.
!
!     RETURN 0.0 IF DENTARG < RZ(MINZ)
!     RETURN KZ  IF DENTARG > RZ(KZ)
!
      INTEGER K
!
      IF     (DENTARG.LT.RZ(MINZ)) THEN
        DENLOC = 0.0
      ELSEIF (DENTARG.EQ.RZ(MINZ)) THEN
        DENLOC = MINZ
      ELSEIF (DENTARG.GE.RZ(KZ)) THEN
        DENLOC = KZ
      ELSE
        DO K= MINZ+1,KZ
          IF     (DENTARG.LE.RZ(K)) THEN !dentarg>rz(k-1)
            DENLOC = K-1 + (DENTARG-RZ(K-1))/(RZ(K)-RZ(K-1))
            EXIT
          ENDIF
        ENDDO !k
      ENDIF
      RETURN
END

SUBROUTINE FIND_DEPTH(ZLOC,ZTARG,Z,KZ,MINZ)
      IMPLICIT NONE

      INTEGER KZ,MINZ
      REAL*4  ZLOC,ZTARG,Z(KZ)
!
!     FIND EXACT LOCATION IN LAYER SPACE OF ZTARG
!     SEARCHING WITHIN Z(MINZ:KZ)
!
!     RETURN 0.0 IF ZTARG < Z(MINZ)
!     RETURN KZ  IF ZTARG > Z(KZ)
!
      INTEGER K
!
      IF     (ZTARG.LT.Z(MINZ)) THEN
        ZLOC = 0.0
      ELSEIF (ZTARG.EQ.Z(MINZ)) THEN
        ZLOC = MINZ
      ELSEIF (ZTARG.GE.Z(KZ)) THEN
        ZLOC = KZ
      ELSE
        DO K= MINZ+1,KZ
          IF     (ZTARG.LE.Z(K)) THEN !ztarg>z(k-1)
            ZLOC = K-1 + (ZTARG-Z(K-1))/(Z(K)-Z(K-1))
            EXIT
          ENDIF
        ENDDO !k
      ENDIF
      RETURN
      END








END MODULe mhycom
