MODULE mprep
USE mkinds
USE mstate
USE mutil
USE mncio
USE mpobs
USE mdates
USE mgrid

CHARACTER(cl)  :: dbLoc
contains

subroutine init_cpobs(cdtg,nobs,pfx)
implicit none
character(len=*) :: cdtg,pfx
integer(i4)      :: k,ivar,ilvl,icount,fid,i,ncount,ig,jg,ii,jj
integer(i4),intent(out) :: nobs
integer(i4)              :: nx,ny
character(:),allocatable       :: filename
real(r4),parameter              :: spval=2.00**100

if(allocated(profile_obs)) deallocate(profile_obs)
ncount=0
filename=trim(tsis_obs_location)//"/tsis_cobs_"//trim(pfx)//"_"//cdtg//".nc"
call nciopn(filename,fid)
call ncioin(filename,fid,"prof_num",ncount)
allocate(profile_obs(nobs_typ,nz,ncount))
call nciorv(filename,fid,"lon",profile_obs%lon)
call nciorv(filename,fid,"lat",profile_obs%lat)
call nciorv(filename,fid,"val",profile_obs%val)
call nciorv(filename,fid,"err",profile_obs%err)
call nciorv(filename,fid,"grdi",profile_obs%gridi)
call nciorv(filename,fid,"grdj",profile_obs%gridj)
call nciorv(filename,fid,"id",profile_obs%id)
call nciocl(filename,fid)
nobs=ncount

!do i=1,ncount
!Write(*,*) "profile No: ", i
!Write(*,*) "***********************************"
! do k=1,nz
!  if(profile_obs(1,k,i)%id==1) then 
!   ii=profile_obs(1,k,i)%gridi
!   jj=profile_obs(1,k,i)%gridj
!   profile_obs(1,k,i)%inov=profile_obs(1,k,i)%val-ftem(ii,jj,k)
!   profile_obs(2,k,i)%inov=profile_obs(2,k,i)%val-fsal(ii,jj,k)
!   profile_obs(4,k,i)%inov=profile_obs(4,k,i)%val-fpin(ii,jj,k)
! write(*,'(9F12.3)') mdepth(ii,jj),profile_obs(1,k,i)%lon,profile_obs(1,k,i)%lat,profile_obs(1,k,i)%val,ftem(ii,jj,k),profile_obs(2,k,i)%val,fsal(ii,jj,k),profile_obs(4,k,i)%val,fpin(ii,jj,k)
!  endif
! enddo
!enddo

end subroutine
        



subroutine init_pobs(cdtg,nobs,pfx)
implicit none
character(len=*) :: cdtg,pfx
integer(i4)      :: k,ivar,ilvl,icount,fid,i,ncount,ig,jg,ii,jj
integer(i4),intent(out) :: nobs
integer(i4)              :: nx,ny
character(:),allocatable       :: filename
real(r4),parameter              :: spval=2.00**100

if(allocated(profile_obs)) deallocate(profile_obs)
ncount=0
filename=trim(tsis_obs_location)//"/tsis_obs_"//trim(pfx)//"_"//cdtg//".nc"
call nciopn(filename,fid)
call ncioin(filename,fid,"prof_num",ncount)
allocate(profile_obs(nobs_typ,nz,ncount))
call nciorv(filename,fid,"lon",profile_obs%lon)
call nciorv(filename,fid,"lat",profile_obs%lat)
call nciorv(filename,fid,"val",profile_obs%val)
call nciorv(filename,fid,"err",profile_obs%err)
call nciorv(filename,fid,"grdi",profile_obs%gridi)
call nciorv(filename,fid,"grdj",profile_obs%gridj)
call nciorv(filename,fid,"id",profile_obs%id)
call nciocl(filename,fid)
nobs=ncount

!do i=1,ncount
!Write(*,*) "profile No: ", i
!Write(*,*) "***********************************"
! do k=1,nz
!  if(profile_obs(1,k,i)%id==1) then 
!   ii=profile_obs(1,k,i)%gridi
!   jj=profile_obs(1,k,i)%gridj
!   profile_obs(1,k,i)%inov=profile_obs(1,k,i)%val-ftem(ii,jj,k)
!   profile_obs(2,k,i)%inov=profile_obs(2,k,i)%val-fsal(ii,jj,k)
!   profile_obs(4,k,i)%inov=profile_obs(4,k,i)%val-fpin(ii,jj,k)
! write(*,'(9F12.3)') mdepth(ii,jj),profile_obs(1,k,i)%lon,profile_obs(1,k,i)%lat,profile_obs(1,k,i)%val,ftem(ii,jj,k),profile_obs(2,k,i)%val,fsal(ii,jj,k),profile_obs(4,k,i)%val,fpin(ii,jj,k)
!  endif
! enddo
!enddo

end subroutine

subroutine pobs2mgrid(nobs,ivar,ilvl,work,err,idx,nx,ny)
implicit none
integer(i4)              :: ivar,ilvl,icount,fid,i,ncount,ig,jg,ii,jj
integer(i4)              :: nx,ny,nobs
real(r4)                 :: work(nx,ny),idx(nx,ny),err(nx,ny)
real(r4),parameter       :: spval=2.00**100
ncount=nobs
work=spval
idx=0.0
do i=1,ncount
  if (profile_obs(ivar,ilvl,i)%id==1) then
   ii=profile_obs(ivar,ilvl,i)%gridi
   jj=profile_obs(ivar,ilvl,i)%gridj  
   work(ii,jj)=profile_obs(ivar,ilvl,i)%val
   err(ii,jj)=profile_obs(ivar,ilvl,i)%err
   idx(ii,jj)=idx(ii,jj)+1.0
 endif
enddo

end subroutine


SUBROUTINE getobs(obsdb,lonbnds,latbnds,timebnds,work_obs,ncount,err,lobs)
IMPLICIT NONE

! Arguments
CHARACTER(Len=*),INTENT(IN)      :: obsdb
REAL(r4),INTENT(IN)              :: lonbnds(2)
REAL(r4),INTENT(IN)              :: latbnds(2)
REAL(r4),INTENT(IN)              :: timebnds(3)
REAL(r4)                         :: err
TYPE(tobs),INTENT(INOUT)         :: work_obs(:)
INTEGER(i4),INTENT(INOUT)        :: ncount
LOGICAL(bl),INTENT(INOUT)          :: lobs
! Local Variables

INTEGER(i4)           :: fid
INTEGER(i4)           :: nvals
INTEGER(i4)           :: index_begin
INTEGER(i4)           :: index_end
INTEGER(i4)           :: k
REAL(r4),allocatable  :: time(:)
REAL(r4),allocatable  :: longitude(:)
REAL(r4),allocatable  :: latitude(:)
REAL(r4),allocatable  :: sla(:)
LOGICAL(bl)           :: obsdb_exists
REAL(r4),parameter    :: taucorr_sla=5.
REAL(r4),parameter    :: fact_sla=1.5
real(r4)              :: tdiff,cntr_time,analysis_time


INQUIRE(FILE=obsdb, EXIST=obsdb_exists)
IF (obsdb_exists) then

call nciopn(trim(obsdb),fid)
call ncioin(trim(obsdb),fid,"time",nvals)
allocate(time(nvals))
allocate(longitude(nvals))
allocate(latitude(nvals))
allocate(sla(nvals))
call nciorv(trim(obsdb),fid,'time',time)
call nciorv(trim(obsdb),fid,'longitude',longitude)
call nciorv(trim(obsdb),fid,'latitude',latitude)
call nciorv(trim(obsdb),fid,'sla_filtered',sla)
call nciocl(trim(obsdb),fid)

longitude=longitude*1.e-6
latitude=latitude*1.e-6
!!Alex sla=sla*0.001  the aviso data have a scale factor of 0.0001 after 2014
sla=sla*0.001
time=time+ndays(01,01,1950,12,31,1900)
where(longitude>180.0) longitude=longitude-360.0
analysis_time=timebnds(3)

!if((timebnds(1)<=time(1) .and. timebnds(2)<=time(1)) .or. (timebnds(1)>=time(nvals) .and. timebnds(2)>=time(nvals))) then
! lobs=.false.
!else
! index_begin=ridx(time,timebnds(1))
! index_end=ridx(time,timebnds(2))
! lobs=.false.
do k=1,nvals
   if(latitude(k)>latbnds(1) .and. latitude(k)<=latbnds(2)) then
           if (longitude(k)>=lonbnds(1) .and. longitude(k)<=lonbnds(2)) then
                   ncount=ncount+1
                   lobs=.true.
                   tdiff=abs(analysis_time-time(k))/taucorr_sla
               work_obs(ncount)%lon=longitude(k)
               work_obs(ncount)%lat=latitude(k)
               work_obs(ncount)%time=time(k)
               work_obs(ncount)%val=sla(k)
               work_obs(ncount)%err=err*(1.0 + fact_sla*(1.0-exp(-tdiff)))
               work_obs(ncount)%lvl=1
!               print *, tdiff,err*(1.0 + fact_sla*(1.0-exp(-tdiff)))

           endif
    endif

enddo
ELSE

!Write(*,*) "File not found:  ", trim(obsdb)        
lobs=.false.
ENDIF
END SUBROUTINE getobs


subroutine ifld(src_nx,src_ny,dst_nx,dst_ny,src_lon,src_lat,src_fld,src_err,dst_fld,dst_err)
USE mkinds
IMPLICIT NONE
INTEGER(KIND=i4)       :: src_nx,src_ny
INTEGER(KIND=i4)       :: dst_nx,dst_ny
! source and destination arrays
REAL(KIND=r4)         :: src_lon(src_nx)
REAL(KIND=r4)         :: src_lat(src_ny)
REAL(KIND=r4)         :: src_fld(src_nx,src_ny)
REAL(KIND=r4)         :: src_err(src_nx,src_ny)
REAL(KIND=r4)         :: dst_fld(dst_nx,dst_ny)
REAL(KIND=r4)         :: dst_err(dst_nx,dst_ny)
! Internal arrays
INTEGER(KIND=i4)       :: xmap(dst_nx,dst_ny,1,4)
INTEGER(KIND=i4)       :: ymap(dst_nx,dst_ny,1,4)
REAL(KIND=r4)          :: wgts(dst_nx,dst_ny,1,4)
! internal variables
INTEGER(KIND=i4)       :: ii
INTEGER(KIND=i4)       :: jj
REAL(KIND=r4)         :: ptlon
REAL(KIND=r4)         :: ptlat
REAL(KIND=r4)         :: grdi
REAL(KIND=r4)         :: grdj
REAL(KIND=r4)         :: r
REAL(KIND=r4)         :: s
REAL(KIND=r4)         :: xsum

DO jj=1,ny
 DO ii=1,nx
         ptlon=mlon(ii,jj)
         ptlat=mlat(ii,jj)
         if(maxval(src_lon)>180.0 .and. ptlon<0.0) ptlon=ptlon+360.0
         grdi=ridx(src_lon,ptlon)
         grdj=ridx(src_lat,ptlat)
         CALL setwgts(src_nx,src_ny,grdi,grdj,xmap(ii,jj,1,:),ymap(ii,jj,1,:),wgts(ii,jj,1,:))
         IF(sum(wgts(ii,jj,:,1)) > 1.01) THEN
                 if(pid==0)    WRITE(*,*)  "Weights Error: ", ii,jj,grdi,grdj
         ENDIF     
 ENDDO
ENDDO
if(pid==0) WRITE(*,*)"Min values of mapping coodinates in OBS Grid: ", src_lat(MINVAL(ymap,mask=ymap.gt.0)),src_lon(MINVAL(xmap,mask=xmap.gt.0))
if(pid==0)WRITE(*,*) "Max values of mapping coodinates in OBS Grid: ", src_lat(MAXVAL(ymap,mask=ymap.gt.0)),src_lon(MAXVAL(xmap,mask=xmap.gt.0))
!    Now map the source grid to destination grid
         CALL  remap(dst_fld(:,:),src_fld(:,:),xmap(:,:,1,:),ymap(:,:,1,:),wgts(:,:,1,:),src_nx,src_ny,dst_nx,dst_ny)
         CALL  remap(dst_err(:,:),src_err(:,:),xmap(:,:,1,:),ymap(:,:,1,:),wgts(:,:,1,:),src_nx,src_ny,dst_nx,dst_ny)
end subroutine
END MODULE mprep
