module msroif
use mkinds
use mncio

!  define parameters, variables, for the ROIF assimilation scheme.

! information pair:

      integer(i4),parameter              :: maxW=2  ! width of MRF neighborhood (rectangular).
      integer(i4),parameter              :: maxV=1  ! number of dynamically independent layers.
      real(r4),target,allocatable        :: infoMatrix(:,:,:,:)  ! info.matrix.
      real(r4),target,allocatable        :: infoVector(:,:)
      real(r4),parameter                 :: vardd = 0.04*0.04
      real(r4),parameter                 :: varbb = 0.125*0.125
      real(r4),parameter                 :: weightD=(1.0/vardd)
      real(r4),parameter                 :: weightB=(1.0/varbb)

      

!      common/roifInfoPair/infoMatrix,infoVector

! sampled statistics:
!      integer maxZl,nEns
!      parameter (maxZl=2)  ! number of vertical levels of samples.
!      parameter (nEns=1)  ! ensemble dimensions.

! variables set in "roifdat.f":
!      real*8  roifUpdateFq
      real(r4)  ::  weightInitialize, weightPredict  ! model weights.
      real(r4)  ::  weightGeostrophy, weightLid, weightDiagonal  ! GeoThin parameters.
      real(r4)  ::  roifTheta(maxV)  ! diagonal of ROIF "Theta"
      real(r4)  ::  scaleTLMu, scaleTLMv, scaleTLMh  ! uncertainties for each TLM eq.
      !real(r4)  ::  varSSH, weightH  ! SSH assimilation variances and weights.
      integer(i4),save ::  maxIteration

! typical local variables:
      integer(i4) ::  i,j,k,l,mw,nw,iw,jw,iww,jww  ! array/matrix indices.


contains

subroutine setInfoPair(idm,jdm,bo,vv,mask)
      ! sets up the information pair for the test.

      implicit none
       integer(i4)                        :: idm,jdm
      ! information pair:
      real(r4),pointer ::  z(:,:)
      real(r4),pointer ::  W(:,:,:,:)
      
      ! local variables:
      real(r4) wThin,r,c
      integer(r4) ia,ja
      real(r4)                     :: bo(idm,jdm),vv(idm,jdm)
      integer(i4)                  :: mask(idm,jdm)
      real(r4),parameter           :: spval=2.00**100
      character(:),allocatable     :: fname
      ! typical local variables:
      integer(i4) ::  i,j,k,l,mw,nw,iw,jw,iww,jww  ! array/matrix indices.

      allocate(infoMatrix(idm,jdm,-maxW:maxW,-maxW:maxW))
      allocate(infoVector(idm,jdm))
      W=>infoMatrix;z=>infoVector
      infoMatrix=0.0
      infoVector=0.0
      c=1.0
      r=0.05
      wThin=1.*weightB
      maxiteration=25

       do j=1,jdm
        do i=1,idm
         if(mask(i,j)==1) then

                 wThin=1./(0.5*vv(i,j)*vv(i,j)+0.01)
        W(i,j,0,0)=20.*wThin


        W(i,j,1,0)=-8.*wThin
        W(i,j,0,1)=-8.*wThin
        W(i,j,0,-1)=-8.*wThin
        W(i,j,-1,0)=-8.*wThin


        W(i,j,1,1)=2.*wThin
        W(i,j,-1,1)=2.*wThin
        W(i,j,1,-1)=2.*wThin
        W(i,j,-1,-1)=2.*wThin


        W(i,j,2,0)=1.*wThin
        W(i,j,0,2)=1.*wThin
        W(i,j,0,-2)=1.*wThin
        W(i,j,-2,0)=1.*wThin

         endif
       enddo
      enddo

        do j=1,jdm
         do i=1,idm
            if(bo(i,j)<spval) then
                z(i,j)=z(i,j)+bo(i,j)*weightD
                W(i,j,0,0)=W(i,j,0,0)+weightD
            endif
          enddo
        enddo  


        do j=1,jdm
         do i=1,idm
          if(W(i,j,0,0).eq.0.) W(i,j,0,0)=1.0
         enddo
        enddo

end subroutine

subroutine roifPCG(po,idm,jdm,mask)
      ! a Preconditioned Conjugate Gradient solution for the
      ! information pair.


      implicit none
       integer(i4)                        :: idm,jdm
      real(r4) :: po(idm,jdm)
      integer(i4) :: mask(idm,jdm)

      ! information pair:
      real(r4),pointer ::  b(:,:),A(:,:,:,:)
      !equivalence (b,infoVector),(A,infoMatrix)
      ! print flag:
      logical diagPCG
      parameter (diagPCG=.true.)  ! if set, prints iteration residuals.

      ! local variables:
      real(r4) ::  x(idm,jdm),r(idm,jdm),p(idm,jdm)
      real(r4),target,allocatable ::  z(:,:)
      real(r4),pointer  ::  w(:,:)
      real(r4) ::  pw,alphaPCG,betaPCG,rhoNew,rhoOld
      integer(i4) ::  iteration,io,jo
      ! typical local variables:
      integer(i4) ::  i,j,k,l,mw,nw,iw,jw,iww,jww  ! array/matrix indices.

      allocate(z(idm,jdm))

       b=>infoVector;A=>infoMatrix
       w=>z

       p=0.0


        do j=1,jdm
         do i=1,idm
         if(mask(i,j)==1) then
           x(i,j)=0.
           r(i,j)=b(i,j)
         endif  
         enddo
        enddo

      do iteration=1,maxIteration

        rhoNew=0.
        do j=1,jdm
         do i=1,idm
         if(mask(i,j)==1) then

         !z(i,j,iw,k)=r(i,j,iw,k)  ! no preconditioning.
          ! preconditioning:
          if(abs(A(i,j,0,0)).gt.1.) then
            z(i,j)=r(i,j)/A(i,j,0,0)
          else
            z(i,j)=r(i,j)
          endif
          ! preconditioning done.
          rhoNew=rhoNew+r(i,j)*z(i,j)
         endif
          enddo
        enddo


      if(iteration.gt.1) betaPCG=rhoNew/rhoOld
        do j=1,jdm
        do i=1,idm
        if(mask(i,j)==1) then

          if(iteration.eq.1) then
            p(i,j)=z(i,j)
          else
            p(i,j)=z(i,j)+betaPCG*p(i,j)
          endif
          w(i,j)=0.
         endif
          enddo
        enddo


        do j=1,jdm
        do i=1,idm
              if(mask(i,j)==1) then

              do nw=-maxW,maxW
                jo=j+nw
                if((jo.ge.1).and.(jo.le.jdm)) then
                  do mw=-maxW,maxW
                    io=i+mw
                   if((io.ge.1).and.(io.le.idm)) then
                      if(A(i,j,mw,nw).ne.0.) then

      w(i,j)=w(i,j)+A(i,j,mw,nw)*p(io,jo)

                      endif
                    endif
                  enddo
                endif
        enddo
         

        endif
        enddo
        enddo

   pw=0.
        do j=1,jdm
        do i=1,idm
        if(mask(i,j)==1) then

          pw=pw+p(i,j)*w(i,j)
        endif
        enddo
        enddo



        alphaPCG=rhoNew/pw
        do j=1,jdm
        do i=1,idm
        if(mask(i,j)==1) then

          x(i,j)=x(i,j)+alphaPCG*p(i,j)
          r(i,j)=r(i,j)-alphaPCG*w(i,j)
        endif
          enddo
        enddo
        
        rhoOld=rhoNew

        if(diagPCG) write(6,90) iteration,rhoOld,alphaPCG,betaPCG
90      format('  PCG ',I6,': ',3G12.3)

      enddo ! iteration

      do j=1,jdm
      do i=1,idm
      if(mask(i,j)==1) then

        po(i,j)=x(i,j)
       endif
        enddo
      enddo

      where(abs(po)<0.01) po=0.0

      end

!--------------------------------------------------
      subroutine statInfo(idm,jdm)

              implicit none

               integer(i4)                        :: idm,jdm
      real(r4) rave,rmax,rr
      ! typical local variables:
      integer(i4) ::  i,j,k,l,mw,nw,iw,jw,iww,jww  ! array/matrix indices.

      ! information pair:
      real(r4),pointer  :: z(:,:),W(:,:,:,:)
      integer(i4)   :: nmax,mmax,imax,jmax

      z=>infoVector
      W=>infomatrix

      nmax=0
      mmax=0
      imax=0
      jmax=0
        rmax=0.
        rave=0.
        do mw=-maxW,maxW
        do nw=-maxW,maxW
          do i=1,idm
          do j=1,jdm
            rr=abs(W(i,j,mw,nw))
            if(rmax.lt.rr) then
              rmax=rr
              imax=i
              jmax=j
              mmax=mw
              nmax=nw
            endif
            rave=rave+rr
          enddo
          enddo
        rave=rave/real(25*idm*jdm)
      write(*,99) rmax,imax,jmax,mmax,nmax,rave 
99      format('W',G12.3,4I3,G12.3)
      enddo
      enddo
      end
    
!subroutine writeROIF(po,idm,jdm)

!        implicit none
!        integer(i4) :: idm,jdm
!        ! typical local variables:
!      integer(i4) ::  i,j,k,l,mw,nw,iw,jw,iww,jww  ! array/matrix indices.

!        real(r4) :: po(idm,jdm),p1(idm,jdm)
!        real(r4) :: xc(idm),yc(jdm)

!        do i=1,idm
!           xc(i)=i
!        enddo

!        do j=1,jdm
!           yc(j)=j
!           enddo


!ncfile="roifOutG.nc"
!create file
!call nciocf(ncfile,fid)

!create dimension and coordinate variable 1
!vname="xc"
!vtype=NF90_Float
!vdim1d(1)=1
!call nciodd(ncfile,fid,vname,idm)
!call nciocv(ncfile,fid,vname,vtype,vdim1d)

!create dimension and coordinate variable 2
!vname="yc"
!vtype=NF90_Float
!vdim1d(1)=2
!call nciodd(ncfile,fid,vname,jdm)
!call nciocv(ncfile,fid,vname,vtype,vdim1d)


!create data variable 4
!vname="po"
!vtype=NF90_Float
!vdim2d=(/1,2/)
!call nciocv(ncfile,fid,vname,vtype,vdim2d)

!create data variable 4
!vname="p1"
!vtype=NF90_Float
!vdim2d=(/1,2/)
!call nciocv(ncfile,fid,vname,vtype,vdim2d)




!call ncioed(ncfile,fid)

! write data into variables

!call nciowv(ncfile,fid,"xc",xc)
!call nciowv(ncfile,fid,"yc",yc)
!call nciowv(ncfile,fid,"po",po)
!call nciowv(ncfile,fid,"p1",p1)
! close file

!call nciocl(ncfile,fid)
!end subroutine
end module msroif      
